"""Mobility Model.

Mobility modeling
"""
from typing import Optional

import confuse  # type: ignore
import numpy as np  # type: ignore # noqa: F401
import pandas as pd  # type: ignore # noqa: F401

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


class Mobility(Base):
    """Governs how we act.

    This contains
    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    Uses a modified standard project
    Uses https://www.sphinx-doc.org/en/master/ to generate the documentation
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Optional[Log] = None,
        type: Optional[str] = None,
    ):
        """Initialize the Economy object.

        This uses the Frame object and populates it with default data unless yo
        override it
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        super().__init__(log_root=log_root)
        log = self.log

        self.mobility_mM_pr: Optional[Data] = None
        log.debug(f"{self.mobility_mM_pr=}")

        self.mob_to_pop_map_mp_pm: Optional[Data] = None



"""Filter Model.

Filter the model to what's interesting. This is what Dashboard can call on when
users are done filtering.
"""

import numpy as np  # type: ignore # noqa: F401
import pandas as pd  # type: ignore # noqa: F401

from .base import Base  # type: ignore
from .log import Log  # type: ignore


# TODO: This will be much easier once we add geo to it
class Filter(Base):
    """Governs how we act.

    This contains
    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    """

    def __init__(
        self,
        log_root: Log = None,
        county: str = None,
        state: str = None,
        subpop: str = None,
    ):
        """Initialize the Economy object.

        This uses the Frame object and populates it with default data unless
        you override it
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        super().__init__(log_root=log_root)

        log = self.log

        self.location = {"county": county, "state": state}
        log.debug(f"{self.location=}")
        self.subpop = subpop
        log.debug(f"{self.subpop=}")



"""Population class.

This is the base class for all models. It instantiates the population to be
filled in.

It also includes a default() for setting up the main attribute matrix,
detail_pd_arr and then a calc() which calculates the rest of the related data
"""

from typing import Optional

# Insert the classes of data we support here
import confuse  # type: ignore

# Note that pip install data-science-types caused errors
from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


class Population(Base):
    """Population objects are created here.

    It has a default model in it for testing which is the Bharat model
    You should override it with a new child class

    Population statistics and model for population
    Initially this contains population of p x 1
    Later it will be p x d where d are the detail columns
    For instance the number of covid patients
    The number of trips or visits or runs for a given population

    The second matrix p population describes is how to map population to:w

    l demand levels to give a p x l. Long term this becomes d x p x l
    So that you can have a different run rate for each detail d of population

    How are resources consumed by different levels in a population
    This is the key first chart in the original model.
    It takes a set of l protection levels and then for each of n resources,
    provides their burn rate. So it is a dataframe that is l x n

    In this first version, burn rates are per capita, that is per person in a
    given level.

    In a later version, we will allow different "burn rates" by population
    attributes so this becomes a 3 dimensional model. For convenience, the
    Frame object we have retains objects in their simple dataframe form since
    it is easy to extract

    For multidimenstional indices, we keep both the n-dimensional array
    (tensor) and also have a method ot convert it to a multiindex for use by
    Pandas

    There is a default mode contained here for testing, you should override
    this by creating a child class and overriding the init

    We also create a friendly name and long description as document strings
    eventually this will become a file we read in that is a data description
    but for now it is a dictionary
    """

    def __init__(self, config: confuse.Configuration, log_root: Log = None):
        """Initialize all variables.

        All initialization here and uses type to determine which method to call
        The default is PopulationDict which reads from the model.data
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        # to pick up the description
        super().__init__(log_root=log_root)
        log = self.log
        log.debug("In %s", __name__)

        self.config = config

        # these need to be filled out by the subclasses
        # define them here for type checking purposes
        # And to have them instantiated for subclasses
        self.population_pP_tr: Optional[Data] = None
        self.pop_demand_per_unit_map_pd_um: Optional[Data] = None
        self.pop_to_popsum1_per_unit_map_pp1_us: Optional[Data] = None



"""Organization Model using Dictionary Input.

Organization modeling
"""
import confuse  # type: ignore

from .data import Data  # type: ignore
from .log import Log  # type: ignore
from .organization import Organization  # type: ignore


class OrganizationDict(Organization):
    """Stub for org from reading a dictionary."""

    def __init__(self, config: confuse.Configuration, log_root: Log = None):
        """Set some variables."""
        super().__init__(config, log_root=log_root)
        log = self.log

        self.organization_oO_tr = Data(
            "organization_oO_tr", config, log_root=log_root
        )

        self.org_demand_per_unit_map_od_um = Data(
            "org_demand_per_unit_map_od_um", config, log_root=log_root
        )

        self.org_to_orgsum1_per_unit_map_oo1_us = Data(
            "org_to_orgsum1_per_unit_map_oo1_us", config, log_root=log_root
        )

        log.debug(f"{self.organization_oO_tr=}")



"""Demand Model.

Demand modeling
"""
from typing import Dict, Optional

import confuse  # type: ignore
import numpy as np  # type: ignore

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore
from .population import Population  # type: ignore
from .resourcemodel import Resource  # type: ignore


class Demand(Base):
    """Calculate consumption based on Population and Resource.

    Take in Pop and and Res and into model.data["Pop Res Demand pn"]
    Some parameters are to use:
    - Washington estimate
    - Mitre burn rates
    - Johns Hopkins burn rates
    - CDPH estimates
    - Ensemble

    If pop and res aren't set by default take the existing Population and
    resource already in the model.data

    With dimensions ["Population p"]["Resource n"]

    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    Uses a modified standard project
    Uses https://www.sphinx-doc.org/en/master/ to generate the documentation
    """

    def __init__(
        self,
        config: confuse.Configuration,
        res: Resource,
        pop: Optional[Population] = None,
        log_root: Optional[Log] = None,
        type: Optional[str] = None,
    ):
        """Initialize the Economy object.

        This uses the Frame object and populates it with default data unless yo
        override it
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        super().__init__(log_root=log_root)
        log = self.log
        log.debug(f"In {__name__}")

        self.set_data(config, log_root)
        self.pop = pop
        self.res = res

        if pop is not None:
            self.recalc()

    def adjust_burn(self, new_burn):
        """Adjust the burn rates."""
        self.demand_per_unit_map_dn_um.array = new_burn
        self.recalc()

    def set_data(self, config, log_root):
        """Instantiate the base data objects for inheritance purposes."""
        log = self.log

        self.demand_per_unit_map_dn_um = Data(
            "demand_per_unit_map_dn_um", config, log_root=log_root
        )
        log.debug(f"{self.demand_per_unit_map_dn_um.df=}")

        self.demand_by_pop_per_person_pn_uc = Data(
            "demand_by_pop_per_person_pn_uc", config, log_root=log_root
        )

        # need to instantiate variables first

        self.demand_by_pop_total_pn_tc = Data(
            "demand_by_pop_total_pn_tc", config, log_root=log_root
        )

        # not used
        self.demand_by_popsum_ps_us: Optional[Dict] = None

        self.demand_by_popsum1_per_person_p1n_uc = Data(
            "demand_by_popsum1_per_person_p1n_uc", config, log_root=log_root
        )

        self.demand_by_popsum1_total_p1n_tc = Data(
            "demand_by_popsum1_total_p1n_tc", config, log_root=log_root
        )
        # get a range enabled version of the above for inventory
        self.demand_by_popsum1_total_rp1n_tc = Data(
            "demand_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )

        self.demand_by_popsum1_total_cost_p1n_xc = Data(
            "demand_by_popsum1_total_cost_p1n_xc", config, log_root=log_root
        )
        log.debug(f"{self.demand_by_popsum1_total_cost_p1n_xc.df=}")

    def recalc(self):
        """Recalculate all demands.

        Right now it must be run in this order
        """
        self.set_demand_by_pop_per_person_pn_uc()
        self.set_demand_by_pop_total_pn_tc()
        self.set_demand_by_popsum1_per_person_p1n_uc()
        self.set_demand_by_popsum1_total_p1n_tc()
        self.set_demand_by_popsum1_total_cost_p1n_xc()

    def set_demand_by_pop_per_person_pn_uc(self):
        """Sets the Demand by Population Per Person."""
        log = self.log

        # original matrix multiply to get per person demands
        # self.demand_pn_arr = np.array(pop.level_pm_df) @ np.array(
        #     self.level_to_res_mn_df
        # )

        if self.pop.pop_demand_per_unit_map_pd_um is None:
            raise ValueError("pop.pop_demand_per_unit_map_pd_um")

        self.demand_by_pop_per_person_pn_uc.array = (
            self.pop.pop_demand_per_unit_map_pd_um.array
            @ self.demand_per_unit_map_dn_um.array
        )
        log.debug(f"{self.demand_by_pop_per_person_pn_uc.df=}")
        # Einsum equivalent for automatic generation
        test = np.einsum(
            "pd,dn->pn",
            self.pop.pop_demand_per_unit_map_pd_um.array,
            self.demand_per_unit_map_dn_um.array,
        )
        log.debug(f"{test=}")
        if not np.array_equal(self.demand_by_pop_per_person_pn_uc.array, test):
            log.critical("einsum fails!")

    def set_demand_by_pop_total_pn_tc(self):
        """Recalcs the Demand by Population for Resources."""
        # Note there is a big hack here as we should really calculate
        # demand across many parameters, but we just pick size
        # self.total_demand_pn_arr = (
        #   np.array(self.demand_pn_df).T * np.array(pop.detail_pd_df["Size"])
        # ).T
        # uses a double transpose because broadcast
        # https://numpy.org/doc/stable/user/basics.broadcasting.html
        # broadcast means start from the lowest dimension
        # each dimenion must match from left to right of be one
        # so if we have pn = (2,9) and p = (2,) we need to invert
        # them to be np=(9,2) and p =(1,2) note that 2, must be converted
        # to a real column matrix
        # Also the size only is just a vector so you need newaxis to promote it
        # https://stackoverflow.com/questions/29241056/how-does-numpy-newaxis-work-and-when-to-use-it
        # need this to prevent type check problems
        log = self.log
        if self.pop.population_pP_tr is None:
            raise ValueError("pop.population_pP_tr")
        self.demand_by_pop_total_pn_tc.array = (
            self.demand_by_pop_per_person_pn_uc.array.T
            * self.pop.population_pP_tr.df["Size"].to_numpy().T
        ).T
        log.debug(f"{self.demand_by_pop_total_pn_tc.df=}")
        test = np.einsum(
            "pn,p->pn",
            self.demand_by_pop_per_person_pn_uc.array,
            self.pop.population_pP_tr.df["Size"].to_numpy(),
        )
        log.debug(f"{test=}")
        if not np.array_equal(self.demand_by_pop_total_pn_tc.array, test):
            breakpoint()
            log.critical("einsum fails!")

    def set_demand_by_popsum1_per_person_p1n_uc(self):
        """Recalcs the Demand by Population Summary Level 1 for Resources."""
        log = self.log
        # Original math put the level or popsum1 data here now move to
        # population
        # self.level_demand_ln_df = np.array(self.level_pl_df).T @ np.array(
        #     self.demand_pn_df
        # )
        if self.pop.pop_to_popsum1_per_unit_map_pp1_us is None:
            raise ValueError("pop.pop_to_popsum1_per_unit_map_pp1_us")
        # then add to them

        self.demand_by_popsum1_per_person_p1n_uc.array = (
            self.pop.pop_to_popsum1_per_unit_map_pp1_us.array.T
            @ self.demand_by_pop_per_person_pn_uc.array
        )
        log.debug(f"{self.demand_by_popsum1_per_person_p1n_uc.df=}")
        # Einsum equivalent of the above, we use x since index needs to be a
        # single character
        test = np.einsum(
            "px,pn->xn",
            self.pop.pop_to_popsum1_per_unit_map_pp1_us.array,
            self.demand_by_pop_per_person_pn_uc.array,
        )
        log.debug(f"{test=}")
        if not np.array_equal(
            self.demand_by_popsum1_per_person_p1n_uc.array, test
        ):
            log.debug("einsum fails!")

    def set_demand_by_popsum1_total_p1n_tc(self):
        """Recalcs the Demand by Population Level 1 for Total Resource."""
        log = self.log
        # TODO: Eventually we will want to calculate this iteration
        # across all summaries so p -> p1 -> p2...
        # We need a general function that knows how to walk
        # a series of transforms so you feed it any point and it knows how to
        # get to any other level
        # Original math
        # self.level_total_demand_ln_df = (
        #    self.level_pl_df.T @ self.total_demand_pn_df
        # )
        self.demand_by_popsum1_total_p1n_tc.array = (
            self.pop.pop_to_popsum1_per_unit_map_pp1_us.array.T
            @ self.demand_by_pop_total_pn_tc.array
        )
        log.debug(f"{self.demand_by_pop_per_person_pn_uc.df=}")
        test = np.einsum(
            "px,pn->xn",
            self.pop.pop_to_popsum1_per_unit_map_pp1_us.array,
            self.demand_by_pop_total_pn_tc.array,
        )
        if not np.array_equal(self.demand_by_popsum1_total_p1n_tc.array, test):
            log.debug("einsum fails f{self.demand_by_popsum1_total_p1n_tc=}")

        log.debug("for future calculate a dummy range version")
        # https://stackoverflow.com/questions/32171917/copy-2d-array-into-3rd-dimension-n-times-python
        # repeat along the first axis
        self.demand_by_popsum1_total_rp1n_tc.array = np.repeat(
            self.demand_by_popsum1_total_p1n_tc.array[np.newaxis, ...],
            self.demand_by_popsum1_total_rp1n_tc.array.shape[0],
            axis=0,
        )

    def set_demand_by_popsum1_total_cost_p1n_xc(self):
        """Recalc the total cost of demand by Population Level 1."""
        # log = self.log
        # original formula
        # self.level_total_cost_ln_df = (
        #       self.level_total_demand_ln_df * cost_ln_df.values)
        log = self.log
        if self.res.res_by_popsum1_cost_per_unit_p1n_us is None:
            raise ValueError("res_by_popsum1_cost_per_unit_p1n_us")
        self.demand_by_popsum1_total_cost_p1n_xc.array = (
            self.demand_by_popsum1_total_p1n_tc.array
            * self.res.res_by_popsum1_cost_per_unit_p1n_us.array
        )
        # TODO: the einsum is broken
        test = np.einsum(
            "xn,xn->xn",
            self.demand_by_popsum1_total_p1n_tc.array,
            self.res.res_by_popsum1_cost_per_unit_p1n_us.array,
        )
        if not np.array_equal(
            self.demand_by_popsum1_total_cost_p1n_xc.array, test
        ):
            log.critical(
                f"einsum fail f{self.demand_by_popsum1_total_p1n_tc=}"
            )



"""Utilities.

Main utilities
"""
import datetime
import math
import os
from pathlib import Path
from typing import Dict, Optional, Union

import bqplot  # type: ignore
import confuse  # type: ignore
import ipysheet  # type: ignore
import ipywidgets as widgets  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from bqplot import pyplot as plt  # type: ignore
from IPython.display import display  # type: ignore

chart_colors = [
    "#77AADD",
    "#99DDFF",
    "#44BB99",
    "#BBCC33",
    "#AAAA00",
    "#EEDD88",
    "#EE8866",
    "#FFAABB",
    "#DDDDDD",
    "#000000",
]


def set_config(path: str):
    """Set a confuse configuration."""
    os.environ["CONFIGDIR"] = os.path.abspath(path)
    config = confuse.Configuration("config")
    return config


def is_dir_or_file(name: str) -> bool:
    """Is path a directory or a file.

    It's hard to believe this is not a function already
    """
    path = Path(name)
    if path.is_dir() or path.is_file():
        return True
    return False


# sets the frame properly but does need to understand the model
# so goes into the model method
def set_dataframe(
    arr: np.ndarray,
    label: Optional[Dict],
    index: Optional[str] = None,
    columns: Optional[str] = None,
) -> pd.DataFrame:
    """Set the dataframe up.

    Using the model data Dictionary and labels
    """
    # we use get so that if there is no item it returns None
    # https://www.tutorialspoint.com/python/dictionary_get.htm
    df = pd.DataFrame(
        arr,
        index=label[index]
        if label is not None and index is not None
        else None,
        columns=label[columns]
        if label is not None and columns is not None
        else None,
    )
    df.index.name = index
    df.columns.name = columns
    return df


def load_dataframe(fname: str) -> pd.DataFrame:
    """Load h5 file into a dataframe.

    Args:
        Name of h5 file

    Returns:
        The dataframe serialized in the h5 file
    """
    df: pd.DataFrame = pd.read_hdf(fname, "df")
    return df


def datetime_to_code(code: Union[str, datetime.datetime]) -> str:
    """Convert datetime objects to valid OCC codes.

    Gets around the problem of Excel automatically converting date-looking
    strings into datetime objects that can't be undone.

    Args:
        code: Either a datetime object or string represnting an OCC code

    Returns:
        The code in valid OCC code format
    """
    if type(code) is datetime.datetime:
        return str(code.month) + "-" + str(code.year)  # type: ignore
    else:
        return str(code)


def to_df(sheet):
    """Shorter function call for sheet -> df."""
    return ipysheet.pandas_loader.to_dataframe(sheet)


def to_sheet(df):
    """Shorter function call for df -> sheet."""
    return ipysheet.pandas_loader.from_dataframe(df)


def format_cells(sheet, money=False):
    """Format ipysheet cells with specific attributes."""
    for cell in sheet.cells:
        setattr(cell, "read_only", True)
        if money is True:
            setattr(cell, "numeric_format", "$0,000")
        else:
            setattr(cell, "numeric_format", "0,000")


def format_population(sheet, money=False, round=False):
    """Generate a formatted sheet optimized for displaying population."""
    df = to_df(sheet)
    if round:
        df = df.round()
    index_name = "Population"
    headers = list(df.index)
    df.insert(loc=0, column=index_name, value=headers)
    sheet = to_sheet(df)
    format_cells(sheet, money)
    sheet.row_headers = False
    return sheet


def display_population(sheet, money=False, round=False):
    """Display sheet with specific, population-optimized formatting."""
    sheet = format_population(sheet, money=money, round=round)
    display(sheet)


def generate_pie_chart(df, title="", show_decimal=False):
    fig = plt.figure(title=title)

    pie_chart = plt.pie(
        sizes=df.values.tolist(),
        labels=df.index.values.tolist(),
        display_labels="outside",
        colors=chart_colors[: df.index.values.size],
        display_values=True,
    )
    if not show_decimal:
        pie_chart.values_format = "0"
    return fig


def generate_bar(df, title="", scientific_notation=False, small_xlabel=False):
    fig = plt.figure(title=title)
    x_vals = df.index.values.tolist()
    if len(x_vals) > 5:
        small_xlabel = True
    x_titles = []
    for val in x_vals:
        if len(val.split(" ")) < 3:
            x_titles.append(val)
        else:
            x_titles.append(" ".join(val.split(" ")[:2]))
    bar_chart = plt.bar(
        x=x_titles, y=df, colors=chart_colors[: df.index.values.size]
    )
    if small_xlabel:
        fig.axes[0].tick_style = {"font-size": "6"}
    if not scientific_notation:
        fig.axes[1].tick_format = ".1f"
    return fig


def generate_group_bar(df, title="", scientific_notation=False):
    fig = plt.figure(title=title)
    bar_chart = plt.bar(
        x=df.columns.values.tolist(),
        y=df,
        labels=df.index.values.tolist(),
        display_legend=False,
        type="grouped",
        colors=chart_colors[: df.index.values.size],
    )
    if df.columns.name:
        plt.xlabel(df.columns.name.rsplit(" ", 1)[0])
    plt.ylim(0, np.amax(df.values))
    if not scientific_notation:
        fig.axes[1].tick_format = ".1f"
    return fig


def generate_scatter(
    df, title="", scientific_notation=False, small_xlabel=True
):
    fig = plt.figure(title=title)
    x_vals = df.index.values.tolist()
    if len(x_vals) > 5:
        small_xlabel = True
    x_titles = []
    for val in x_vals:
        if len(val.split(" ")) < 3:
            x_titles.append(val)
        else:
            x_titles.append(" ".join(val.split(" ")[:2]))
    scatter = plt.scatter(x=x_titles, y=df)

    if small_xlabel:
        fig.axes[0].tick_style = {"font-size": "6"}
    if not scientific_notation:
        fig.axes[1].tick_format = ".1f"
    return fig


def generate_stacked_bar(df, title="", scientific_notation=False):
    fig = plt.figure(title=title)

    bar_chart = plt.bar(
        x=df.columns.values.tolist(),
        y=df,
        labels=df.index.values.tolist(),
        display_legend=False,
        type="stacked",
        colors=chart_colors[: df.index.values.size],
    )
    if df.columns.name:
        plt.xlabel(df.columns.name.rsplit(" ", 1)[0])
    plt.ylim(0, np.amax(df.values))
    if not scientific_notation:
        fig.axes[1].tick_format = ".1f"
    return fig


def generate_separate_bar_list(
    df, scientific_notation=False, small_xlabel=False
):  # returns list, NOT widget
    bar_list = []
    for col in df.columns:  # .values.tolist()
        bar_list.append(
            generate_bar(
                df[col][df[col] != 0],
                title=col,
                scientific_notation=scientific_notation,
                small_xlabel=small_xlabel,
            )
        )
    return bar_list


def generate_separate_scatter_list(
    df, scientific_notation=False, small_xlabel=False
):  # returns list, NOT widget
    scatter_list = []
    for col in df.columns:  # .values.tolist()
        scatter_list.append(
            generate_scatter(
                df[col][df[col] != 0],
                title=col,
                scientific_notation=scientific_notation,
                small_xlabel=small_xlabel,
            )
        )
    return scatter_list


def generate_html_legend(df, colors=chart_colors, table=True, font_size=13):
    name = df.index.name.rsplit(" ", 1)[0]
    html_string = f"<div style='font-size:{font_size}px; font-family:helvetica'><b style='font-weight:bold'>{name}</b>"
    indices = df.index.values
    if table:
        html_string += "<table><tr>"
        for i in range(0, indices.size):
            if i % 2 == 0:
                index_num = int(i / 2)
            else:
                index_num = int(i / 2 + indices.size / 2)
            html_string += f"<td style='padding:0 5px'><span style='color:{colors[index_num]}'>█</span> {indices[index_num]}</td>"
            if i % 2 != 0:
                html_string += "</tr><tr>"
        html_string += "</tr></table>"
    else:
        for_count = 0
        for string in indices:
            if for_count == 0:
                html_string += "<br>"
            else:
                html_string += "&emsp;"
            html_string += (
                f"<span style='color:{colors[for_count]}'>█</span> {string}"
            )
            for_count += 1
    html_string += "</div>"
    return widgets.HTML(html_string)


def generate_group_bar_legend(
    df, title="", scientific_notation=False, legend_table=True
):
    chart = generate_group_bar(
        df, title=title, scientific_notation=scientific_notation
    )
    legend = generate_html_legend(df, table=legend_table)
    return widgets.VBox([chart, legend])


def generate_stacked_bar_legend(
    df, title="", scientific_notation=False, legend_table=True
):
    chart = generate_stacked_bar(
        df, title=title, scientific_notation=scientific_notation
    )
    legend = generate_html_legend(df, table=legend_table)
    return widgets.VBox([chart, legend])


def triangular(a, b, c):
    return math.sqrt(((a * a + b * b + c * c) - a * b - a * c - b * c) / 18)



"""Mobility Data Read from a large table.

Read in the mobility from an external table
"""

import confuse  # type: ignore
import pandas as pd  # type: ignore # noqa: F401

from .data import Data  # type: ignore
from .log import Log  # type: ignore
from .mobility import Mobility  # type: ignore


class MobilityTable(Mobility):
    """Mobility Data Readers.

    Reads the mobility data. The default is to read from the model.data
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
        type: str = None,
    ):
        """Initialize the mobility object.

        This uses the Frame object and populates it with default data unless
        you override it
        """
        # the base class handles log management
        super().__init__(config, log_root=log_root)
        # convenience name for log
        log = self.log

        if config is None:
            raise ValueError(f"{config=} is null")

        # get mobility data
        # Using the new Lucas data class
        self.mobility_mM_pr = Data(
            "mobility_mM_pr",
            config,
            log_root=log_root,
        )
        log.debug(f"{self.mobility_mM_pr.df=}")



"""Compose the model without a CLI."""
import os
import pathlib
from typing import Optional

import numpy as np  # type: ignore

from restart.log import Log
from restart.model import Model
from restart.util import set_config

PATH = pathlib.Path(__file__).parent.absolute()


class RestartModel:
    """Bootstrap a model object in a notebook environment."""

    def __init__(
        self,
        population: str = "dict",
        organization: Optional[str] = None,
        csv: Optional[str] = None,
        county: Optional[str] = None,
        state: Optional[str] = None,
        subpop: Optional[str] = None,
        config: str = "default",
        output: Optional[str] = None,
        resource: str = "dict",
        inventory: str = "dict",
        demand: str = "dict",
        financial: str = "dict",
        mobility: str = "ensemble",
        epi: str = "imhe",
    ):
        """Initialize a model object."""
        # set up all attributes
        for i in locals().items():
            if i[0] != "self":
                setattr(self, i[0], i[1])

        # set the logging
        self.name = "model"
        self.log_root = Log(self.name)
        self.log = log = self.log_root.log
        log.propagate = False
        log.debug(f"{__name__=}")

        # set up the config
        self.config = set_config(os.path.join(PATH, f"config/{config}"))
        self.set_model()

    def recalc_burn(self, burn: np.ndarray):
        """Change a model attribute and recalculate."""
        self.model.demand.adjust_burn(burn)

    def set_model(self, **kwargs):
        """Bootstrap the model."""
        # override defaults with keywords
        for k, v in kwargs.items():
            setattr(self, k, v)
        # build the model
        self.model = (
            Model(self.name, log_root=self.log_root)
            .set_configure(self.config)
            .set_filter(
                county=self.county, state=self.state, subpop=self.subpop
            )
            .set_population(type=self.population)
            .set_organization(type=self.organization)
            .set_resource(type=self.resource)
            .set_inventory(type=self.inventory)
            .set_demand(type=self.demand)
            .set_financial(type=self.financial)
            .set_epi(type=self.epi)
            .set_mobility(type=self.mobility)
            .set_output(out=self.output, csv=self.csv)
        )

        self.demand = self.model.demand



"""Base class for Loader.

Base class
"""
from typing import Dict, Optional

from .base import BaseLog  # type: ignore
from .log import Log  # type: ignore


class Load(BaseLog):
    """Base Load YAML Files.

    Base configuration from YAML files
    """

    # no variable unless they are the same across all instances

    def __init__(
        self,
        *paths,
        log_root: Optional[Log] = None,
    ):
        """Initialize Loader Base Class.

        Base class just sets a logger
        """
        super().__init__(log_root=log_root)
        log = self.log
        self.data: Dict = {}

        # replace the standalone logger if asked
        log = self.log

        log.debug(f"{self.log=} {log=}")
        log.debug(f"module {__name__=}")



"""Tester for main model module."""

from restart.restart import RestartModel

if __name__ == "__main__":
    oes_test = RestartModel(population="oes", state="California")
    dict_test = RestartModel(population="dict")



"""The Inventory Model.

Reads from dictionary for defaults
"""

import confuse  # type: ignore

from .inventory import Inventory  # type: ignore
from .log import Log  # type: ignore


class InventoryDict(Inventory):
    """InventoryDict - manages inventory reads from dictionary."""

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
    ):
        """Initialize the inventorys.

        Reads from the default config yaml files
        """
        # The default is read from the dictionary
        super().__init__(config, log_root=log_root)
        log = self.log
        log.debug(f"in {__name__}")



"""Population reading from OES data.

Population is working
"""
import math
from typing import Dict, List, Optional, Tuple

import confuse  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from restart_datasets import data  # type: ignore

from .data import Data  # type: ignore
from .filtermodel import Filter  # type: ignore
from .log import Log  # type: ignore
from .population import Population  # type: ignore
from .util import datetime_to_code  # type: ignore


class PopulationOES(Population):
    """Transforms OES data into a format compatible with the model.

    Performs calculations to give us an estimate of population distributions on
       a county-wide basis.

    Attributes:
        oes_df: Dataframe containing OES data
        code_df: Dataframe containing conversions between county and MSA code
        pop_df: Dataframe containing census population data per county
        cty_name: Name of a US county
        state_name: Name of a US state
        df: The processed, OES data in a dataframe
    """

    def __init__(
        self,
        config: confuse.Configuration,
        filt: Filter,
        data_dir: str = None,
        log_root: Optional[Log] = None,
    ):
        """Initialize.

        Read the paths in and create dataframes, generate mappings
        """
        super().__init__(config, log_root=log_root)
        self.log_root = log_root
        log = self.log

        log.debug(f"module {__name__=}")

        # get location and population from the filter
        self.location = filt.location
        try:
            if self.location["county"] is not None:
                self.location["county"] += " County"
        except KeyError:
            log.debug("invalid location input {self.location=}")
            return

        self.subpop = filt.subpop
        self.codes: list
        self.data_dir = data_dir

        self._load_data(config, self.location)

    def _load_data(self, config, location):
        """Do most of the initializing here.

        That way the stuff we don't want passed is hidden.
        """
        code_df = self._format_code(data.list1_2020())
        pop_df = data.co_est2019_alldata()
        xls_df = self._format_map(data.covid_surge_who())
        # initialize unsliced dataframe from oes data
        if location["county"] is None and location["state"] is None:
            oes_df = data.us()
            df = self._create_country_df(oes_df)
        elif location["county"] is not None and location["state"] is not None:
            oes_df = data.counties()
            df = self._create_county_df(location, oes_df, code_df, pop_df)
        else:
            # oes_df = getattr(data, location["state"])
            st = location["state"].title().replace(" ", "")
            oes_df = eval(f"data.{st}()")  # nosec
            df = self._create_state_df(location, oes_df)

        # filter the population
        if self.subpop == "healthcare":
            df = self._health_filter(df)

        elif self.subpop == "wa_tier2_opt1":
            df = self._wa_tier2_opt1_filter(df)

        elif self.subpop == "wa_tier2_opt2":
            df = self._wa_tier2_opt2_filter(df)

        elif self.subpop == "wa_groupings":
            df = self._wa_public_safety_filter(df)

        # the actual data passed onto the model
        self.pop_detail_df = self._drop_code(df)
        self.population_pP_tr = Data(
            "population_pP_tr",
            config,
            log_root=self.log_root,
            p_index=list(self.pop_detail_df.index),
            P_index=["Size"],
            array=self._drop_code(df).to_numpy(),
        )

        pop_to_burn_df = self._pop_to_burn_rate(df, xls_df)
        self.pop_demand_per_unit_map_pd_um: Data = Data(
            "pop_demand_per_unit_map_pd_um",
            config,
            log_root=self.log_root,
            p_index=list(pop_to_burn_df.index),
            array=pop_to_burn_df.to_numpy(),
        )

        if self.subpop != "wa_groupings":
            self._set_essential(xls_df, config)
        else:
            self.pop_to_popsum1_per_unit_map_pp1_us = Data(
                "pop_to_popsum1_per_unit_map_pp1_us",
                config,
                log_root=self.log_root,
            )

    def _format_code(self, df: pd.DataFrame) -> pd.DataFrame:
        """Perform dataframe transformations specific to list1_2020.xls.

        Args:
            df: A dataframe

        Returns:
            A transformed dataframe to match the format needed for this project
        """
        # specify columns to bypass issues with underlining in original excel
        df.columns = [
            "CBSA Code",
            "MDC Code",
            "CSA Code",
            "CBSA Title",
            "Metropolitan/Micropolitan Statistical Area",
            "Metropolitan Division Title",
            "CSA Title",
            "County Equivalent",
            "State Name",
            "FIPS State Code",
            "FIPS County Code",
            "Central/Outlying County",
        ]

        # Select MSA, as this is what OES data is based off of
        df = df[
            df["Metropolitan/Micropolitan Statistical Area"]
            == "Metropolitan Statistical Area"
        ]

        # Drop data we don't need
        df = df.drop(
            [
                "MDC Code",
                "CSA Code",
                "Metropolitan Division Title",
                "Metropolitan/Micropolitan Statistical Area",
                "CSA Title",
                "FIPS State Code",
                "FIPS County Code",
                "Central/Outlying County",
            ],
            axis=1,
        )

        # Reset indices for aesthetic appeal
        df = df.reset_index(drop=True)

        return df

    def _format_map(self, df: pd.DataFrame) -> pd.DataFrame:
        """Manually slice the excel model to get protection level mappings.

        Args:
            df: The excel model loaded into a dataframe

        Returns:
            The dataframe sliced to give the mappings
        """
        # manually redo indexing and select the rows we need
        DF_COLUMNS = 2528
        DF_START = 2529
        DF_END = 3303

        df.columns = df.iloc[DF_COLUMNS]
        df = df.iloc[DF_START:DF_END]
        df = df[
            ["Washington SOT", "SOC", "Type", "Level", "Essential (0 lowest)"]
        ]

        # fix datetime objects and drop empty rows
        df["SOC"] = df["SOC"].apply(datetime_to_code)
        df = df.dropna(axis="rows").reset_index(drop=True)
        return df

    def _pop_to_burn_rate(
        self, df: pd.DataFrame, map_df: pd.DataFrame
    ) -> Tuple[list, np.ndarray]:
        """Generate mappings for OCC codes and population levels.

        Args:
            df: A dataframe that has OCC codes

        Returns:
            Dictionary of the population level mappings
        """
        map_arr = []
        labels = []
        for code in df["occ_code"]:
            arr = np.zeros(7)
            try:
                ind = map_df[map_df["SOC"] == code].index[0]
                level = map_df.iloc[ind]["Level"]
            except IndexError:
                if code.startswith("29-") or code.startswith("31-"):
                    level = 5.5
                else:
                    level = 3

            # assign integer levels
            if type(level) is int:
                arr[level] = 1

            # assign multiple levels
            else:
                arr[math.floor(level)] = 0.5
                arr[math.ceil(level)] = 0.5

            # add to dictionary
            name = df[df["occ_code"] == code].index.tolist()[0]
            labels.append(name)
            map_arr.append(arr)

        pop_to_level_df = pd.DataFrame(map_arr, index=labels)

        return pop_to_level_df

    def _find_code(self, location: Dict, code_df: pd.DataFrame) -> int:
        """Finds the MSA code of given county.

        Args:
            None

        Returns:
            Integer corresponding to the given county's MSA code
        """
        if code_df is None:
            raise ValueError(f"{code_df=} should not be None")

        return int(
            code_df[
                (code_df["County Equivalent"] == location["county"])
                & (code_df["State Name"] == location["state"])
            ]["CBSA Code"].iloc[0]
        )

    def _calculate_proportions(
        self,
        code: int,
        location: Dict,
        code_df: pd.DataFrame,
        pop_df: pd.DataFrame,
    ) -> float:
        """Calculate county proportion relative to total MSA pop.

        Args:
            code: MSA code for desired county

        Returns:
            A float corresponding to the ratio of the county's population in
            relation to its MSA code.
        """
        if code_df is None:
            raise ValueError(f"{code_df=} should not be None")
        if pop_df is None:
            raise ValueError(f"{code_df=} should not be None")

        # list the counties in the same MSA code as cty_name
        counties = list(
            code_df[code_df["CBSA Code"] == str(code)]["County Equivalent"]
        )

        # construct dictionary mapping county names to constituent populations
        populations = {}
        for county in counties:
            pop = int(
                pop_df[
                    (pop_df["CTYNAME"] == county)
                    & (pop_df["STNAME"] == location["state"])
                ]["POPESTIMATE2019"]
            )
            populations[county] = pop

        # calculate total population in MSA code
        total_pop = sum(populations.values())

        # divide individual county population by total MSA population
        return populations[location["county"]] / total_pop

    def _load_county(
        self,
        location: Dict,
        oes_df: pd.DataFrame,
        code_df: pd.DataFrame,
        pop_df: pd.DataFrame,
    ) -> Tuple[float, pd.DataFrame]:
        """Slice the OES data by county for further processing downstream.

        Args:
            None

        Returns:
            proportion: Float corresponding to proportion of residents from
                        MSA code living in given county
            df: Sliced OES dataframe
        """
        # find county MSA CODE
        code = self._find_code(location, code_df)

        # calculate proportion of MSA code's residents living in county
        proportion = self._calculate_proportions(
            code, location, code_df, pop_df
        )

        # initialize dataframe as slice of OES data
        df = oes_df[oes_df["area"] == code][
            ["occ_code", "occ_title", "o_group", "tot_emp"]
        ]

        # replace placeholders with 0
        df = df.replace(to_replace="**", value=0)

        return proportion, df

    def _load_state(
        self, location: Dict, oes_df: pd.DataFrame
    ) -> pd.DataFrame:
        """Slice the OES data by state for further processing downstream.

        Args:
            None

        Returns:
            df: Sliced OES dataframe
        """
        # replace placeholders with 0
        df = oes_df.replace(to_replace="**", value=0)
        df.drop(["Unnamed: 0"], axis=1, inplace=True)
        return df

    def _load_country(self, oes_df: pd.DataFrame) -> pd.DataFrame:
        """Get the OES data for the whole country.

        The default setting for OES population
        """
        # slice OES dataframe by the whole county
        col_list = ["occ_code", "occ_title", "o_group", "tot_emp", "naics"]
        df = oes_df[col_list]
        df = df.drop(["naics"], axis=1)
        # replace placeholders with 0
        df = df.replace(to_replace="**", value=0)

        return df

    def _fill_uncounted(
        self, major: pd.DataFrame, detailed: pd.DataFrame
    ) -> pd.DataFrame:
        """Create special categories for uncounted employees.

        Args:
            major: Dataframe containing totals for major OCC categories
            detailed: Dataframe containing totals for detailed OCC categories

        Returns:
            The detailed dataframe with extra categories to account for
            uncounted workers
        """
        code_list = list(set(major["occ_code"]))

        for code in code_list:
            pat = code[0:3]
            filt = detailed[detailed["occ_code"].str.startswith(pat)]

            # calculate number of employees unaccounted for within the major
            # OCC code
            total = int(major[major["occ_code"] == code]["tot_emp"])
            det_total = np.sum(filt.astype({"tot_emp": "int"})["tot_emp"])
            delta = total - det_total

            if delta > 0:
                # create dataframe row and append to detailed dataframe
                name = list(major[major["occ_code"] == code]["occ_title"])[0]
                add_lst = [
                    [pat + "XXXX", "Uncounted " + name, "detailed", delta]
                ]
                add_df = pd.DataFrame(add_lst, columns=list(major.columns))
                detailed = detailed.append(add_df, ignore_index=True)

        return detailed

    def _format_output(self, df: pd.DataFrame) -> pd.DataFrame:
        """Format dataframe to fit the model by dropping some columns.

        Args:
            df: The dataframe we want to format

        Returns:
            The formatted dataframe
        """
        df = df.drop(df[df["tot_emp"] == 0].index)
        df = df.drop(["o_group"], axis=1)
        df = df.reset_index(drop=True)
        return df.astype({"tot_emp": "int"})

    def _drop_code(self, df: pd.DataFrame) -> pd.DataFrame:
        """Drop the OCC code from a dataframe.

        So that it has the right format for the model.
        """
        col_labs = ["Size"]
        self.codes = list(df["occ_code"])
        df = df.drop(["occ_code"], axis=1)
        df.columns = col_labs

        return df

    def _create_county_df(
        self,
        location: Dict,
        oes_df: pd.DataFrame,
        code_df: pd.DataFrame,
        pop_df: pd.DataFrame,
    ) -> pd.DataFrame:
        """Generate dataframe containing processed OES data by county.

        Args:
            None

        Returns:
            The processed dataframe
        """
        # load in sliced dataframe
        proportion, df = self._load_county(location, oes_df, code_df, pop_df)

        # split into 'major' and 'detailed' OCC categories
        major = df[df["o_group"] == "major"].copy()
        detailed = df[df["o_group"] == "detailed"].copy()

        # some detailed categories don't have information availble - remove
        # these and place into "Uncounted" category
        detailed = self._fill_uncounted(major, detailed)

        # adjust 'tot_emp' columns by MSA code proportion
        detailed["tot_emp"] = (
            detailed["tot_emp"]
            .astype("int")
            .apply(lambda x: int(x * proportion))
        )

        # format to fit model
        detailed = self._format_output(detailed)
        detailed.set_index("occ_title", drop=True, inplace=True)
        return detailed

    def _create_state_df(
        self, location: Dict, oes_df: pd.DataFrame
    ) -> pd.DataFrame:
        """Generate dataframe containing processed OES data by state.

        Args:
            None

        Returns:
            The processed dataframe
        """
        # load in sliced dataframe
        df = self._load_state(location, oes_df).drop(["area"], axis=1)
        major = df[df["o_group"] == "major"].copy()
        detailed = df[df["o_group"] == "detailed"].copy()

        # some detailed categories don't have information available - remove
        # these and place into "Uncounted" category
        detailed = self._fill_uncounted(major, detailed)

        # Format to fit model
        detailed = self._format_output(detailed)
        detailed.set_index("occ_title", drop=True, inplace=True)
        return detailed

    def _create_country_df(self, oes_df: pd.DataFrame) -> pd.DataFrame:
        """Generate dataframe containing processed OES data for US.

        Args:
            oes_df: Dataframe containing OES data

        Returns:
            The processed dataframe
        """
        df = self._load_country(oes_df)
        major = df[df["o_group"] == "major"].copy()
        detailed = df[df["o_group"] == "detailed"].copy()
        detailed = self._fill_uncounted(major, detailed)
        detailed = self._format_output(detailed)
        detailed.set_index("occ_title", drop=True, inplace=True)
        return detailed

    def _health_filter(self, df: pd.DataFrame) -> pd.DataFrame:
        """Return a detailed breakdown of healthcare workers with OCC codes.

        Args:
            None

        Returns:
            Dataframe object with the detailed breakdown
        """
        # 29-XXXX and 31-XXXX are the healthcare worker codes
        filt = df[
            (df["occ_code"].str.startswith("29-"))
            | (df["occ_code"].str.startswith("31-"))
        ]
        return filt

    def _wa_tier2_opt1_filter(self, df: pd.DataFrame) -> pd.DataFrame:
        """Return a detailed breakdown of Washington tier 2 workers.

        Args:
            None

        Returns:
            Dataframe object with the detailed breakdown
        """
        filt = df[
            (df["occ_code"].str.startswith("33-"))
            | (df["occ_code"].str.startswith("29-"))
            | (df["occ_code"].str.startswith("31-"))
        ]

        return filt

    def _wa_tier2_opt2_filter(self, df: pd.DataFrame) -> pd.DataFrame:
        """Return a detailed breakdown of Washington tier 2 workers.

        Args:
            None

        Returns:
            Dataframe object with the detailed breakdown
        """
        occ_list = [
            "29-1021",
            "29-1022",
            "29-1023",
            "29-1024",
            "29-1029",
            "29-1041",
            "29-1051",
            "29-1071",
            "29-1122",
            "29-1123",
            "29-1124",
            "29-1125",
            "29-1127",
            "29-1128",
            "29-1292",
            "29-1141",
            "29-1151",
            "29-1161",
            "29-1171",
            "29-1181",
            "29-1211",
            "29-1212",
            "29-1213",
            "29-1214",
            "29-1215",
            "29-1216",
            "29-1217",
            "29-1218",
            "29-1221",
            "29-1222",
            "29-1223",
            "29-1224",
            "29-1229",
            "29-1241",
            "29-1242",
            "29-1243",
            "29-1249",
            "29-2011",
            "29-2012",
            "29-2031",
            "29-2032",
            "29-2033",
            "29-2034",
            "29-2035",
            "29-2036",
            "29-2042",
            "29-2043",
            "29-2052",
            "29-2053",
            "29-2055",
            "29-2056",
            "29-2057",
            "29-2061",
            "29-2099",
            "29-9021",
            "29-9093",
            "29-9099",
            "31-1120",
            "31-9091",
            "31-9092",
            "31-9093",
            "31-9096",
            "31-9097",
            "31-9099",
            "31-1131",
            "31-1132",
            "39-4011",
            "31-1133",
            "31-9091",
            "33-3052",
        ]

        filt = df[df["occ_code"].isin(occ_list)]

        return filt

    def _wa_public_safety_filter(self, df: pd.DataFrame) -> pd.DataFrame:
        """Return a detailed breakdown of Washington public safety."""
        occ_list = [
            "29-2040",
            "33-2011",
            "33-3012",
            "33-3021",
            "33-9093",
            "33-3041",
            "33-3051",
        ]

        filt = df[df["occ_code"].isin(occ_list)]
        return filt

    def _set_wa_sum(self, df: pd.DataFrame, config) -> pd.DataFrame:
        """Washington group summarization."""
        pop_level: List = []  # noqa: F841

    def _set_essential(self, df: pd.DataFrame, config) -> pd.DataFrame:
        """Get population essential levels from the excel model.

        Manually slice the dataframe
        """
        pop_level: List = []
        df["SOC"] = df["SOC"].apply(datetime_to_code)
        df.reset_index(drop=True, inplace=True)

        for code in list(self.codes):
            arr = np.zeros(2)
            try:
                ind = df[df["SOC"] == code].index[0]
            except IndexError:
                ind = -1
            if ind > 0:
                level = df.iloc[ind]["Essential (0 lowest)"]
            else:
                level = np.random.randint(0, high=6)
            if level >= 5:
                arr[0] = 1
            else:
                arr[1] = 1
            pop_level.append(arr)

        self.pop_to_popsum1_per_unit_map_pp1_us = Data(
            "pop_to_popsum1_per_unit_map_pp1_us",
            config,
            log_root=self.log_root,
            p_index=list(self.pop_detail_df.index),
            array=np.array(pop_level),
        )



"""Population Data Read.

Read in the population from the model dictionary
"""

import confuse  # type: ignore
import pandas as pd  # type: ignore # noqa: F401

from .data import Data  # type: ignore
from .log import Log  # type: ignore
from .population import Population  # type: ignore


class PopulationDict(Population):
    """Population Data Readers.

    Reads the population data. The default is to read from the model.data
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
    ):
        """Initialize the population object.

        This uses the Frame object and populates it with default data unless
        you override it
        """
        # the base class handles log management
        super().__init__(config, log_root=log_root)
        # convenience name for log
        log = self.log

        if config is None:
            raise ValueError(f"{config=} is null")

        # get population data
        # Using the new Lucas data class
        self.population_pP_tr = Data(
            "population_pP_tr",
            config,
            log_root=log_root,
        )
        log.debug(f"{self.population_pP_tr=}")

        self.pop_demand_per_unit_map_pd_um = Data(
            "pop_demand_per_unit_map_pd_um",
            config,
            log_root=log_root,
        )
        log.debug(f"{self.pop_demand_per_unit_map_pd_um=}")

        self.pop_to_popsum1_per_unit_map_pp1_us = Data(
            "pop_to_popsum1_per_unit_map_pp1_us",
            config,
            log_root=log_root,
        )
        log.debug(f"{self.pop_to_popsum1_per_unit_map_pp1_us=}")



"""The Resource Model.

What resources are and how they are consumed
https://mypy.readthedocs.io/en/stable/cheat_sheet_py3.html
"""
# allows return self typing to work
from __future__ import annotations

from typing import Optional

import confuse  # type: ignore

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


class Resource(Base):
    """Resource - Manages all the resources that are used in the model.

    This creates for all r resources, the list of attributes a

    This contains
    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    Uses a modified standard project
    Uses https://www.sphinx-doc.org/en/master/ to generate the documentation

    Stores all in lxn levels and items:
        Resource. Attributes of each
        Costs. Of the resource
        Inventory. The current inventory level
        Safety stock. the minimum inventory level in days since we have a surge
        model this is simply right now the daily rate the model shows
        Economic Order Quantity. inv_eoc_ln_df
    """

    # no variables here unless you want class variables the same across all
    # instances

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
    ):
        """Initialize the Resources.

        Does a read in
        """
        # initialize logging and description
        super().__init__(log_root=log_root)
        log = self.log
        self.config = config
        log.debug(f"in {__name__}")

        # Filling these is the job of the child classes
        self.resource_nN_ur: Optional[Data] = None
        self.res_by_popsum1_cost_per_unit_p1n_us: Optional[Data] = None



"""Base class for our own Data.

Since we are doing multity dimensional math, we need to keep three data
structures in sync:

    - The Numpy array that is multiple dimensions
    - A "wide" DataFrame using Multiindex that compresses these to 2-D to be
      used for printing and tabular displa
    - A "narrow" with reset_index and melt that has a single column of data
    and all the rest is encoded in a huge label set
    - It also carries a description, labels and other meta data

The idea is that when the numpy array changes, then we automatically update the
other two representations
"""
from __future__ import annotations

from typing import Tuple

import confuse  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .base import BaseLog  # type: ignore
from .log import Log  # type: ignore


class Data(BaseLog):
    """Base data class for the entire model.

    The main data is in an Numpy array
    The meta data is
        - description. Markdown text describing the data
        - labels. An ordered list that is the axis name and the columns
           for each
                  As an example a 3-D numpy array would have
                  [ Type: "Resource n"
                    Kind: ["N95", "ASTM-3 Mask"]],
                  [ Type: "Attribute a"
                    Kind: ["Unit", "Volume", "Area"]]

    You can override what is in config by passing keyword arguments
    like
        Data(label="Overriding Label"
    The output data is
        - .array numpy array for computation
        - .df Suitable for display
        - .narrow Suitable for graphing

    Uses config for initial configuration but probably that should be a
    subclass of this as config can come in differently


    """

    # do not put variable here unless you want them the same
    # across all classes see https://docs.python.org/3/tutorial/classes.html

    # https://stackoverflow.com/questions/9056957/correct-way-to-define-class-variables-in-python
    def __init__(
        self,
        key,
        config: confuse.Configuration,
        log_root: Log = None,
        **kwargs,
    ):
        """Set base variables.

        Mainly the descriptions
        """
        # Sets logging
        super().__init__(log_root=log_root)
        log = self.log
        log.debug(f"{__name__=}")

        self.key: str = key
        self.config_cf: confuse.Configuration = config

        # Override the YAML with a dictionary
        # https://confuse.readthedocs.io/en/latest/
        # https://www.pythoncentral.io/how-to-check-if-a-list-tuple-or-dictionary-is-empty-in-python/
        if kwargs:
            for k, v in kwargs.items():
                if "index" in k:
                    k = k.split("_")[0]
                    ind = [
                        i
                        for i in config["Model"][self.key]["index"].get()
                        if f"({k})" in i
                    ][0]
                    args = {"Dimension": {ind: v}}
                else:
                    args = {"Model": {self.key: {k: v}}}

                config.set_args(args, dots=True)

        self.data_cf: confuse.Configuration = config["Model"][key]
        self.dimension_cf: confuse.Configuration = config["Dimension"]
        self.index_cf: confuse.Configuration = self.data_cf["index"]

        self._array: np.ndArray = None
        self._df: pd.DataFrame = None
        self._narrow: pd.DataFrame = None

        # if there is no value it is a calculate amount and fill with NaNs
        try:
            self._array = np.array(self.data_cf["array"].get())
        # user create exceptions must include the import module name
        # https://www.geeksforgeeks.org/user-defined-exceptions-python-examples/
        except confuse.NotFoundError:
            log.debug(f"set null array based on {self.index_cf.get()=}")
            shape = [
                len(self.dimension_cf[x].get()) for x in self.index_cf.get()
            ]
            log.debug(f"of {shape=}")
            self._array = np.empty(
                [len(self.dimension_cf[x].get()) for x in self.index_cf.get()]
            )

        log.debug(f"{self._array=}")

        self.set_df()

    # https://www.journaldev.com/22460/python-str-repr-functions
    # so debugging make sense this does not work because
    # if called early _df and others are not available
    # def __repr__(self):
    #     """Show more than just the type."""
    # return {"df": self._df, "array": self._array, "narrow": self._narrow}

    # Three data structures that
    # can change and then the other two must switch:
    # TODO: Instead of all this manually, change so that we have a
    # an array of transforms so we can index in and call a function
    # The call will look like transform = [ unity, to_numpy ], [ to_df, unity
    # ]]
    # and then the generic function is convert[from][to]
    # - array. This is the numpy array
    # - df. This is the wide format dataframe suitable for printing
    # - narrow. This is the narrow frame for charting
    # Each setter when called must active the other two
    # these are the functions called when something externally is changed
    @property
    def array(self):
        """Property method for array.

        Returns the private variable
        """
        return self._array

    @array.setter
    def array(self, array: np.ndarray):
        """Change display and graph values when array changes.

        When this changes update other representations such as the df and wide
        """
        if self.config_cf is None:
            raise ValueError(f"{self.config_cf=} is null")
        self._array = array
        self.set_df()
        # note set_narrow assumes set_df is done first
        self.set_narrow()

    @property
    def df(self):
        """Property method for df.

        Returns the private variable
        """
        return self._df

    @df.setter
    def df(self, df: pd.DataFrame) -> Data:
        """Change the array and narrow when df changes."""
        self._df = df
        self.set_array(self._df)
        self.set_narrow()
        return self

    @property
    def narrow(self):
        """Property method for narrow form used for charting."""
        return self._narrow

    # TODO: Not implemented should rebuild the indices
    @narrow.setter
    def narrow(self, narrow: pd.DataFrame):
        self._narrow = narrow
        return self

    # set functions called internally after external changes made
    def set_array(self, df: pd.DataFrame):
        """Change the array when df changes."""
        self._array = np.array(df)

    def set_df(self):
        """Change the df when array changes in self._array.

        The normal 2D case when len(dimensions) <=2
        dimension[0] or index is the list of row names
            (e.g., ["Healthcare", "Non-healthcare"])
        dimension[1] or columns is the list of column names
            (e.g., ["N95", "Surgical Mask"])
        index[0] or index_name is name of index  (e.g., "Population")
        index[1] column_name is the name of all the columns (e.g., "Resource")

        The 3D cases we use a multiindex when len(dimensions) >2
        index is a list of a list of row names  for the 1..N-1 dimensions
            [["Min Inv", "EOQ Inv", "Order Size"], ["HC", "Non-HC"]]
        dimension[-1] or columns is the value of the last dimension
        index[-1] or column_name is the same just for the Nth index
        dimension[:-1] or the 1st to N-1st dimensions
        index[:-1] index_name is a list ["Inv", "Population"]

        Note that if you change the data value, the only way to do it appears
        to be to construct a new DataFrame
        https://note.nkmk.me/en/python-pandas-numpy-conversion/
        """
        # Assume that once we set the Data we never change the
        # dimensions or lables, but only the data inside
        self.index_name = self.index_cf.get()[:-1]
        self.columns_name = self.index_cf.get()[-1]
        self.columns = self.dimension_cf[self.columns_name].get()
        if len(self.index_name) <= 1:
            # this is two dimensional
            # https://stackoverflow.com/questions/18691084/what-does-1-mean-in-numpy-reshape
            # Also note that we need .get for a confus object
            # And need to dereference the index_name
            self.index_name = self.index_name[0]
            self.index = self.dimension_cf[self.index_name].get()
            df = pd.DataFrame(
                self.array,
                index=self.index,
                columns=self.columns,
            )
            df.index.name = self.index_name
        else:
            # flatten the array into 2D so only column is there
            # first -1 means compute it, second -1 means last dimension
            self.flat2d_arr = self.array.reshape(-1, self.array.shape[-1])
            # Do a lookup of the index in dimension
            # https://stackoverflow.com/questions/49542348/using-a-list-comprehension-to-look-up-variables-works-with-globals-but-not-loc/49542378
            self.index = [self.dimension_cf[x].get() for x in self.index_name]
            self.multi_index = pd.MultiIndex.from_product(
                self.index,
                names=self.index_name,
            )
            # https://docs.heliopy.org/en/0.1b5/examples/multiindex.html
            df = pd.DataFrame(
                self.flat2d_arr,
                index=self.multi_index,
                columns=self.columns,
            )
        df.columns.name = self.columns_name
        self._df = df

    def set_narrow(self):
        """Convert self.wide to self.narrow."""
        narrow = self.df.reset_index()
        narrow = narrow.melt()
        self._narrow = narrow

    def __iter__(self):
        """Iterate over all Pandas DataFrames.

        Uses a list of all frames
        """
        self.df_list = [
            k for k, v in vars(self).items() if isinstance(v, pd.DataFrame)
        ]
        self.df_len = len(self.df_list)
        self.df_index = 0
        return self

    def __next__(self) -> Tuple[str, pd.DataFrame]:
        """Next Pandas DataFrame.

        Iterates through the list of dataframes
        """
        if self.df_index >= self.df_len:
            raise StopIteration
        key = self.df_list[self.df_index]
        value = vars(self)[key]
        self.df_index += 1
        return key, value



"""Epi model.

The disease model work.
"""

# https://mypy.readthedocs.io/en/stable/cheat_sheet_py3.html
# note the noqa: and type: are space sensitive
# https://stackoverflow.com/questions/51179109/set-pyflake-and-mypy-ignore-same-line
from typing import Optional

import confuse  # type: ignore
import numpy as np  # type: ignore # noqa: F401
import pandas as pd  # type: ignore # noqa: F401

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


class Epi(Base):
    """Resource - Manages all the resources that are used in the model.

    This creates for all r resources, the list of attributes a

    This contains
    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    Uses a modified standard project
    Uses https://www.sphinx-doc.org/en/master/ to generate the documentation
    """

    # no variable here unless you want them the same across all classes
    # see https://docs.python.org/3/tutorial/classes.html
    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Optional[Log] = None,
        type: Optional[str] = None,
    ):
        """Initialize the Epi object.

        This uses the Frame object and populates it with default data unless yo
        override it
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        super().__init__(log_root=log_root)
        log = self.log

        self.epi_eE_pr: Optional[Data] = None
        log.debug(f"{self.epi_eE_pr=}")



"""Epi Data Read.

Read in the epi from the model dictionary
"""
import os
from typing import Dict, List, Optional, Tuple

import confuse  # type: ignore
import h5py  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore # noqa: F401

from .data import Data  # type: ignore
from .epi import Epi  # type: ignore
from .log import Log  # type: ignore


class EpiTable(Epi):
    """Epi Data Readers.

    Reads the epi data. The default is to read from the model.data
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Optional[Log] = None,
        type: str = None,
    ):
        """Initialize the epi object.

        This uses the Frame object and populates it with default data unless
        you override it
        """
        # the base class handles log management
        super().__init__(config, log_root=log_root)
        # convenience name for log
        log = self.log

        if config is None:
            raise ValueError(f"{config=} is null")

        arr: np.ndarray = self.load_and_slice(config)

        # get epi data
        # Using the new Lucas data class
        self.epi_eE_pr = Data(
            "epi_eE_pr",
            config,
            log_root=self.log_root,
            e_index=["COVID-19"],
            E_index=["Cases", "Hospitalizations", "Deaths"],
            array=arr,
        )

        log.debug(f"{self.epi_eE_pr.df=}")

    def load_and_slice(self, config: confuse.Configuration) -> np.ndarray:
        """Builds array of targets for the specified parameters.

        Loads data and indices of cube of Reich Lab collected forecasts,
        slices cube based on parameters, and returns target array (arr).

        Targets include number of new cases during week ending on specified
        date, number of new deaths during week ending on specified date,
        and number of new hospitalizations on day of specified date. Targets
        pertain to COVID-19 only (for now).

        An entry in the target array with a value of -1 indicates that
        that particular target for the specified parameters was not included
        in the Reich Lab collected forecasts.
        """

        def get_state(fips: int) -> (int):
            strfips: str = str(fips)
            if len(strfips) == 4:
                strfips = "0" + strfips
            return int(strfips[:2])

        def get_state_county(fips: int) -> (Tuple[int, int]):
            fips_str: str = str(fips)
            # check if loc is a state
            if len(fips_str) <= 2:
                return fips, 0
            # loc is a county
            state: int = get_state(fips)
            s: slice = slice(start=len(str(state)), stop=len(fips_str))
            county: int = int(fips_str[s])
            return state, county

        # set parameters for now, will take as input soon
        fips: int = 53  # WA state fips
        date: pd.Timestamp = pd.to_datetime("2020-09-05")
        state: Optional[int] = None
        county: Optional[int] = None
        state, county = get_state_county(fips)
        quantile: float = 0.5
        model: str = "IHME-CurveFit"

        source: Dict[str, str] = config["Reich"].get()

        arr: np.ndarray = None

        # read in data cube and corresponding index labels
        with h5py.File(
            os.path.join(source["Root"], source["CUBE"]), "r"
        ) as cube:
            indexdf: pd.DataFrame = pd.read_csv(
                os.path.join(source["Root"], source["INDEX"]), index_col=0
            )

            shape: Tuple[int, int, int, int, int, int] = cube[
                "zipped_reichcube"
            ].shape

            # prune indices of nans and compute index values from parameters
            date_indices: List[pd.Timestamp] = list(
                pd.to_datetime(indexdf["date_index"][: shape[0]])
            )
            state_indices: List[int] = list(
                indexdf["state_index"][: shape[1]].astype(int)
            )
            county_indices: List[int] = list(
                indexdf["county_index"][: shape[2]].astype(int)
            )
            quantile_indices: List[float] = list(
                indexdf["quantile_index"][: shape[3]]
            )
            model_indices: List[str] = list(indexdf["model_index"][: shape[4]])
            d: int = date_indices.index(date)
            s: int = state_indices.index(state)
            c: int = county_indices.index(county)
            q: int = quantile_indices.index(quantile)
            m: int = model_indices.index(model)

            # slice cube with indices to obtain array of targets
            # ":" as last index is because we want all three targets
            arr = cube["zipped_reichcube"][d, s, c, q, m, :]

            # reverse sparse rep and set proper shape
            arr[arr == 0] = -2
            arr[arr == -1] = 0
            arr[arr == -2] = -1
            arr = arr.reshape((1, shape[5]))

        return arr



"""The Inventory Model.

What inventory are and how they are consumed
"""
# allows return self typing to work
from __future__ import annotations

# For slices of parameters
from enum import Enum
from typing import List

import confuse  # type: ignore
import numpy as np  # type: ignore

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


# https://docs.python.org/3/library/enum.html
# These are the slices used
class InvParam(Enum):
    """List positions in Inventory Parameter List."""

    INIT = 0
    EOQ = 1
    MIN = 2


class Inventory(Base):
    """Inventory - Manages all the inventorys that are used in the model."""

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
    ):
        """Initialize the Inventorys.

        Does a read in parameters
        """
        # initialize logging and description
        super().__init__(log_root=log_root)
        log = self.log
        self.config = config
        log.debug(f"in {__name__}")

        self.inv_by_popsum1_total_rp1n_tc = Data(
            "inv_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )

        self.inv_by_popsum1_param_iIp1n_tp = Data(
            "inv_by_popsum1_param_iIp1n_tp", config, log_root=log_root
        )
        log.debug(f"{self.inv_by_popsum1_param_iIp1n_tp.df=}")

        # TODO: This should be taken from the param file
        self.inv_init_by_popsum1_total_rp1n_tc = Data(
            "inv_init_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )
        log.debug(f"set inv to {self.inv_init_by_popsum1_total_rp1n_tc=}")
        self.inv_eoq_by_popsum1_total_rp1n_tc = Data(
            "inv_eoq_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )
        self.inv_min_by_popsum1_total_rp1n_tc = Data(
            "inv_min_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )

        # Helpers to handle period calculations
        self.inv_min_by_popsum1_in_periods_rp1n_pc = Data(
            "inv_min_by_popsum1_in_periods_rp1n_pc", config, log_root=log_root
        )
        self.inv_average_orders_by_popsum1_per_period_rp1n_uf = Data(
            "inv_average_orders_by_popsum1_per_period_rp1n_uf",
            config,
            log_root=log_root,
        )

        self.inv_order_by_popsum1_total_rp1n_tc = Data(
            "inv_order_by_popsum1_total_rp1n_tc", config, log_root=log_root
        )

        # can only set minmum once inv_min exists and order too
        self.set_min(self.inv_init_by_popsum1_total_rp1n_tc)

    def set_average_orders_per_period(
        self, inv_average_orders_by_popsum1_per_period_rp1n_uf: Data
    ):
        """Set Average Inventory Used Every Period.

        This could just be a simple set but leave here for clarity
        """
        self.inv_average_orders_by_popsum1_per_period_rp1n_uf = (
            inv_average_orders_by_popsum1_per_period_rp1n_uf
        )

    def set_min_in_periods(
        self,
        min_periods_r_pc: List,
    ) -> Inventory:
        """Sets the Minimum Inventory as measured in Average Days Shipments.

        A Helper function that fill out an entire array and then passes it
        down
        """
        log = self.log
        log.debug(f"{min_periods_r_pc=}")
        self.inv_min_by_popsum1_in_periods_rp1n_pc.array = np.einsum(
            "r,rxn->rxn",
            min_periods_r_pc,
            np.ones_like(self.inv_min_by_popsum1_in_periods_rp1n_pc.array),
        )
        self.set_min_in_periods_array(
            self.inv_min_by_popsum1_in_periods_rp1n_pc
        )
        return self

    def set_min_in_periods_array(
        self,
        min_periods_rp1n_tc: Data,
    ) -> Inventory:
        """Set with an array that is for all resources in periods."""
        log = self.log
        self.inv_min_by_popsum1_in_periods_rp1n_pc.array = (
            min_periods_rp1n_tc.array
        )
        log.debug(f"{self.inv_min_by_popsum1_in_periods_rp1n_pc.df=} ")

        # https://numpy.org/doc/stable/reference/generated/numpy.empty_like.html
        # note we need r=1 for this to work so we insert an empty dimension
        # https://numpy.org/doc/stable/reference/generated/numpy.expand_dims.html
        # needed this before we started calling with full range
        # self.inv_min_by_popsum1_per_period_rp1n_uc.array = np.expand_dims(
        #     self.inv_min_by_popsum1_per_period_rp1n_uc.array, axis=0
        # )
        # need to do a dot product
        self.inv_min_by_popsum1_total_rp1n_tc.array = np.einsum(
            "rxn,rxn->rxn",
            min_periods_rp1n_tc.array,
            self.inv_average_orders_by_popsum1_per_period_rp1n_uf.array,
        )
        self.set_min(self.inv_min_by_popsum1_total_rp1n_tc)
        return self

    def set_min(self, min_by_popsum1_total_rp1n_tc: Data) -> Inventory:
        """Set the minimum inventory in periods_r.

        This sets the minimum inventory and then forces an order in case we are
        below the minimum
        """
        log = self.log

        # https://stackoverflow.com/questions/53375161/use-numpy-array-to-replace-pandas-dataframe-values
        self.inv_by_popsum1_total_rp1n_tc.array = (
            min_by_popsum1_total_rp1n_tc.array
        )
        log.debug(f"{self.inv_by_popsum1_total_rp1n_tc.df=}")
        self.supply_order()
        return self

    def supply_order(self) -> Inventory:
        """Order from supplier.

        Order up to the minimum inventory
        """
        # hack here because we only do ranges for min inventory
        self.inv_order_by_popsum1_total_rp1n_tc.array = (
            self.inv_min_by_popsum1_total_rp1n_tc.array
            - self.inv_by_popsum1_total_rp1n_tc.array
        )
        # negative means we have inventory above safety levels
        # so get rid of those
        # https://www.w3inventory.com/python-exercises/numpy/python-numpy-exercise-90.php
        self.inv_order_by_popsum1_total_rp1n_tc.array[
            self.inv_order_by_popsum1_total_rp1n_tc.array < 0
        ] = 0
        # now gross up the order to the economic order quantity
        self.round_up_to_eoq(self.inv_order_by_popsum1_total_rp1n_tc)
        self.log.debug(f"{self.inv_order_by_popsum1_total_rp1n_tc.df=}")

        # now that we have an order rounded up and ready, let's get supply
        self.fulfill(self.inv_order_by_popsum1_total_rp1n_tc)
        return self

    # https://stackoverflow.com/questions/2272149/round-to-5-or-other-number-in-python
    def round_up_to_eoq(self, order_by_popsum1_total_rp1n_tc: Data) -> Data:
        """Round order up the economic order quantity.

        Each order needs to get rounded up to an economic quantity
        """
        if np.any(self.inv_eoq_by_popsum1_total_rp1n_tc.array <= 0):
            raise ValueError(
                f"EOQ not positive {self.inv_eoq_by_popsum1_total_rp1n_tc.df=}"
            )

        if np.any(order_by_popsum1_total_rp1n_tc.array < 0):
            raise ValueError(
                f"Negative order in {order_by_popsum1_total_rp1n_tc.df=}"
            )

        # So take the order and then get the distance to the eoc
        # by using modulo
        # https://stackoverflow.com/questions/50767452/check-if-dataframe-has-a-zero-element
        # https://numpy.org/doc/stable/reference/generated/numpy.any.html
        # https://softwareengineering.stackexchange.com/questions/225956/python-assert-vs-if-return
        # do not use asserts they are stripped with optimization, raise errors
        return (
            order_by_popsum1_total_rp1n_tc.array
            + (
                self.inv_eoq_by_popsum1_total_rp1n_tc.array
                - order_by_popsum1_total_rp1n_tc.array
            )
            % self.inv_eoq_by_popsum1_total_rp1n_tc.array
        )

    def fulfill(self, order_by_popsum1_total_rp1n_tc: Data):
        """Fulfill an order form supplier.

        This is a stub in that all orders are immediatley fulfilled
        """
        log = self.log
        log.debug(f"fulfill {order_by_popsum1_total_rp1n_tc=}")
        self.inv_by_popsum1_total_rp1n_tc.array += (
            order_by_popsum1_total_rp1n_tc.array
        )
        log.debug(f"{self.inv_by_popsum1_total_rp1n_tc.df=}")

    def order(self, order_by_popsum1_total_rp1n_tc: Data) -> Data:
        """Order by Customer from Inventory.

        Take a new order and then return what you can

        It will check what is in inventory and then call the delivery method
        returns: whats available to ship
        """
        # Return as much as we can so if the order is bigger than
        # the inventory, just ship it all out.
        # the simple min won't work, need an element0-wise minimum
        # https://numpy.org/doc/stable/reference/generated/numpy.minimum.html
        self.inv_order_by_popsum1_total_rp1n_tc.array = np.minimum(
            order_by_popsum1_total_rp1n_tc.array,
            self.inv_by_popsum1_total_rp1n_tc.array,
        )

        # ship it!
        self.inv_by_popsum1_total_rp1n_tc.array -= (
            self.inv_order_by_popsum1_total_rp1n_tc.array
        )

        # now restock
        self.supply_order()
        return self.inv_order_by_popsum1_total_rp1n_tc



"""Organization Model.

Organization modeling
"""
from typing import Optional

import confuse  # type: ignore

from .base import Base  # type: ignore
from .data import Data  # type: ignore
from .log import Log  # type: ignore


class Organization(Base):
    """The amount of use by Organization.

    This contains
    This uses https://realpython.com/documenting-python-code/
    docstrings using the NumPy/SciPy syntax
    Uses a modified standard project
    Uses https://www.sphinx-doc.org/en/master/ to generate the documentation
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Optional[Log] = None,
        type: Optional[str] = None,
    ):
        """Initialize the Organization object.

        This uses the Frame object and populates it with default data unless yo
        override it
        """
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        # pass the logger down
        super().__init__(log_root=log_root)
        self.config = config
        # create a sublogger if a root exists in the model
        # self.log_root = log_root
        # log = self.log = (
        #     log_root.log_class(self)
        #     if log_root is not None
        #     else logging.getLogger(__name__)
        # )
        # the sample code to move up the logging for a period and then turn it
        # off
        log = self.log
        log.debug(f"in {__name__=}")

        self.organization_oO_tr: Optional[Data] = None
        self.org_demand_per_unit_map_od_um: Optional[Data] = None
        self.org_to_orgsum1_per_unit_map_oo1_us: Optional[Data] = None



"""Model definition.

The model shape is configured here.
And this uses chained methods as decorators
https://www.w3schools.com/python/python_classes.asp
"""
# https://stackoverflow.com/questions/33533148/how-do-i-specify-that-the-return-type-of-a-method-is-the-same-as-the-class-itsel
# this allows Model to refer to itself
from __future__ import annotations

from typing import Generator, List, Optional, Tuple

from .base import Base  # type: ignore
from .demand import Demand  # type: ignore
from .demand_dict import DemandDict  # type: ignore
from .epi import Epi  # type: ignore
from .epi_dict import EpiDict  # type: ignore
from .epi_table import EpiTable  # type: ignore
from .filtermodel import Filter  # type: ignore
from .financial import Financial  # type: ignore
from .financial_dict import FinancialDict  # type: ignore
from .financial_table import FinancialTable  # type: ignore
from .inventory import Inventory  # type: ignore
from .inventory_dict import InventoryDict  # type: ignore
from .log import Log  # type: ignore
from .mobility import Mobility  # type: ignore
from .mobility_dict import MobilityDict  # type: ignore
from .mobility_table import MobilityTable  # type: ignore
from .organization import Organization  # type: ignore
from .organization_dict import OrganizationDict  # type: ignore
from .output import Output  # type: ignore
from .population import Population  # type: ignore
from .population_dict import PopulationDict  # type: ignore
from .population_oes import PopulationOES  # type: ignore
from .population_wa import PopulationWA  # type: ignore
from .resource_dict import ResourceDict  # type: ignore
from .resourcemodel import Resource  # type: ignore


class Model(Base):
    """Main model for planning.

    They then will create the correct tables for use by the main computation
    This model has pointers to the major elements of the model.

    The model is made up of a series of classes for each of the major model
    elements. Each of the major model elements are based on a class structure
    that are kept in a network of links in the Model class. The Model acts as a
    single global data structure for the entire project.

    It uses chaining so that you can in a single statement set all the modeling
    elements. These include:

    Real resources. Like Populations or Organizations
    Transforms. Which computes mapping of Population onto say Resources
    Actions. These are things that affect Real Objects like Demand.

    - LogBase. Just for logging so every class that wants to log needs this as
       a base class
    - Base. This has the descriptions baked in. Used to traverse the Model
      class
    when you want to print or interrogate the Model.
    - Resource. Every time you create a new way to read or manage resources,
      base on this. en.utf-8.add
    """

    # https://satran.in/b/python--dangerous-default-value-as-argument
    # https://stackoverflow.com/questions/2…

    # do not do default assignment, it remembers it on eash call
    # https://docs.python.org/3/library/typing.html
    def __init__(self, name, log_root: Optional[Log] = None):
        """Initialize the model."""
        # the long description of each
        # https://stackoverflow.com/questions/1385759/should-init-call-the-parent-classs-init/7059529
        super().__init__(log_root=log_root)
        log = self.log
        log.debug(f"{__name__=}")

        self.name: str = name
        if not log.hasHandlers():
            print(f"{log=} has no handlers")
        log.debug(f"{self.name=}")

    def set_configure(self, config) -> Model:
        """Configure the Model.

        Uses Loaded as a dictionary and puts it into model variables
        """
        log = self.log
        self.config = config
        log.debug(f"{self.config=}")

        return self

    # TODO: This should be a generated set of methods as they are all identical
    def set_population(self, type: str = None, data_dir: str = None) -> Model:
        """Create population class for model.

        Population created here
        """
        # the old method
        # self.population = Population(
        #    self.data, log_root=self.log_root, type=type
        # )
        # the super class population uses type to return the exact model
        # filter is by happens after this
        self.population: Population
        self.filter: Filter

        if type == "oes":
            self.population = PopulationOES(
                self.config,
                self.filter,
                data_dir=data_dir,
                log_root=self.log_root,
            )
        elif type == "wa":
            self.population = PopulationWA(
                self.config, self.filter, log_root=self.log_root
            )
        elif type == "dict":
            # change this to the the naming of columns
            self.population = PopulationDict(
                self.config,
                log_root=self.log_root,
            )
        else:
            raise ValueError(f"{type=} not implemented")
        return self

    def set_organization(self, type: str = None) -> Model:
        """Set organization."""
        self.organization: Organization
        if type == "dict":
            self.organization = OrganizationDict(
                self.config, log_root=self.log_root
            )

        return self

    def set_resource(self, type: str = None) -> Model:
        """Create resource class.

        Resource
        """
        self.resource: Resource
        if type == "dict":
            self.resource = ResourceDict(self.config, log_root=self.log_root)
        return self

    def set_inventory(self, type: str = None) -> Model:
        """Create Inventory management for a specific warehouse."""
        self.inventory: Inventory
        if type == "dict":
            self.inventory = InventoryDict(self.config, log_root=self.log_root)
        return self

    def set_demand(self, type: str = None) -> Model:
        """Set demand.

        Demand by population levels l
        """
        log = self.log
        self.demand: Demand
        if type == "mitre":
            log.debug("Use Mitre demand")
            raise ValueError("{type=} not implemented")
        elif type == "jhu":
            log.debug("Use JHU burn rate model")
            raise ValueError("{type=} not implemented")
        else:
            log.debug("Use default yaml dictionary data")
            self.demand = DemandDict(
                self.config,
                res=self.resource,
                pop=self.population,
                log_root=self.log_root,
                type=type,
            )
        return self

    def set_filter(
        self, county: str = None, state: str = None, subpop: str = None
    ) -> Model:
        """Filter the model.

        Shrink the model to relevant population, resource
        """
        self.filter = Filter(
            log_root=self.log_root,
            county=county,
            state=state,
            subpop=subpop,
        )

        return self

    def set_financial(self, type: str = None) -> Model:
        """Create Financial model.

        Financial creation
        """
        log = self.log
        self.financial: Financial
        if type == "dict":
            self.financial = FinancialDict(
                self.config, log_root=self.log_root, type=type
            )
        elif type == "table":
            self.financial = FinancialTable(
                self.config, log_root=self.log_root, type=type
            )
        else:
            log.error(f"Financial Model {type=} not implemented")
        return self

    def set_epi(self, type: str = None) -> Model:
        """Create Epi model.

        Epi create
        """
        log = self.log
        self.epi: Epi
        if type == "dict":
            self.epi = EpiDict(self.config, log_root=self.log_root, type=type)
        elif type in [
            "ihme",
            "delphi",
            "icl",
            "lanl",
            "sikjalpha",
            "yyg",
            "chensemble",
        ]:
            self.epi = EpiTable(self.config, log_root=self.log_root, type=type)
        else:
            log.error(f"Epi Model {type=} not implemented")
        return self

    def set_mobility(self, type: str = None) -> Model:
        """Create Behavior model.

        Behavior create
        """
        log = self.log
        self.mobility: Mobility
        if type == "dict":
            self.mobility = MobilityDict(
                self.config, log_root=self.log_root, type=type
            )
        elif type == "table":
            self.mobility = MobilityTable(
                self.config, log_root=self.log_root, type=type
            )
        else:
            log.error("Behavior not implemented")
        return self

    # https://docs.python.org/3/library/typing.html#typing.Generator
    # returns yield type, send type return type which are null
    def walk(self) -> Generator[Tuple[str, Base], None, None]:
        """Walk through all Base objects in Model.

        This is needed for things like Output which are called by Model
        and you cannot mutually import it. Use a generator instead. Because
        python cannot interpret this correctly.
        """
        log = self.log
        for name, value in self:
            log.debug(f"{name=} {value=}")
            yield name, value

    def set_output(self, out: str = None, csv: str = None) -> Model:
        """Generate output."""
        self.output = Output(
            self.walk,
            config=self.config,
            log_root=self.log_root,
            out=out,
            csv=csv,
        )
        return self

    # https://stackoverflow.com/questions/37835179/how-can-i-specify-the-function-type-in-my-type-hints
    # https://www.datacamp.com/community/tutorials/python-iterator-tutorial
    # https://towardsdatascience.com/how-to-loop-through-your-own-objects-in-python-1609c81e11ff
    # So we want the iterable to be the Base Class
    # The iterator is Model which can return all the Base classes
    # https://thispointer.com/python-how-to-make-a-class-iterable-create-iterator-class-for-it/
    def __iter__(self) -> Model:
        """Iterate through the model getting only Base objects."""
        log = self.log
        self.base_list: List = [
            k for k, v in vars(self).items() if isinstance(v, Base)
        ]
        log.debug(f"{self.base_list=}")
        self.base_len: int = len(self.base_list)
        self.base_index: int = 0
        return self

    def __next__(self) -> Tuple[str, Base]:
        """Next Base."""
        log = self.log
        if self.base_index >= self.base_len:
            raise StopIteration
        log.debug(f"{self.base_index=}")
        key = self.base_list[self.base_index]
        value = vars(self)[key]
        log.debug(f"{key=} {value=}")
        self.base_index += 1
        return key, value

    # Use the decorator pattern that Keras and other use with chaining
    def set_logger(self, name: str = __name__) -> Model:
        """Set Log.

        Setup the root logger and log
        """
        self.log_root = Log(name)
        self.log = self.log_root.log
        return self



"""Population reading from QCEW data."""
import math
import os
from typing import Optional

import confuse  # type: ignore
import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .filtermodel import Filter  # type: ignore
from .load_csv import LoadCSV  # type: ignore
from .log import Log  # type: ignore
from .population import Population  # type: ignore
from .util import load_dataframe  # type: ignore


class PopulationWA(Population):
    """NAICS codes for Washington."""

    def __init__(
        self,
        config: confuse.Configuration,
        filt: Filter,
        log_root: Optional[Log] = None,
    ):
        """Initialize.

        Read the paths in and create dataframes, generate mappings
        """
        self.root_log: Optional[Log]
        # global log
        # log: logging.Logger = logging.getLogger(__name__)

        super().__init__(config, log_root=log_root)
        log = self.log
        log.debug(f"module {__name__=}")

        # get location and population from the filter
        self.location = filt.location
        try:
            if self.location["county"] is not None:
                self.location["county"] += " County"
        except KeyError:
            log.debug("invalid location input {self.location=}")
            return

        self.subpop = filt.subpop
        self.config = config

        source = config["Paths"].get()
        source = LoadCSV(source=source).data
        map_df = self.format_map(
            load_dataframe(os.path.join(source["Root"], source["MAP"]))
        )
        self.get_population(map_df)
        self.create_map(map_df)

    def get_population(self, map_df: pd.DataFrame):
        """Get the population numbers."""
        # select the rows we want
        if map_df is None:
            raise ValueError(f"{map_df=} is null")
        if self.config is None:
            raise ValueError(f"{self.config=} is null")

        df = map_df[
            ["Industry description", "6-digit NAICS", "Average employment"]
        ]
        # select just the 6-digit codes
        df = df[df["6-digit NAICS"].apply(lambda x: len(str(x)) == 6)]
        self.detail_pd_df = df.drop("6-digit NAICS", axis=1)
        self.detail_pd_df.columns = ["Industry description", "Size"]
        self.detail_pd_df.set_index(
            "Industry description", drop=True, inplace=True
        )
        self.detail_pd_arr = np.array(self.detail_pd_df)
        self.set_description(
            f"{self.detail_pd_df=}",
            self.config["Description"]["Population p"]["Pop Detail pd"].get(),
        )
        self.log.debug(f"{self.description['detail_pd_df']=}")

    def format_map(self, df: pd.DataFrame) -> pd.DataFrame:
        """Manually slice the excel model to get protection level mappings.

        Args:
            df: The excel model loaded into a dataframe

        Returns:
            The dataframe sliced to give the mappings
        """
        # manually redo indexing and select the rows we need
        # TODO: un-hardcode
        cols = df.iloc[672].dropna()
        df.columns = df.iloc[672]
        df = df.iloc[674:1937]
        df = df[cols]
        df = df.replace(to_replace="*", value=1)

        return df

    def create_map(self, map_df: pd.DataFrame):
        """Generate mappings for NAICS codes and population levels.

        Args:
            df: A dataframe that has NAICS codes

        Returns:
            Dictionary of the population level mappings
        """
        df = map_df[
            [
                "Industry description",
                "6-digit NAICS",
                "Average employment",
                "Protection level",
            ]
        ]

        detailed = df[df["6-digit NAICS"].apply(lambda x: len(str(x)) == 6)]
        medium = df[df["6-digit NAICS"].apply(lambda x: len(str(x)) == 3)]
        major = df[
            df["6-digit NAICS"].apply(
                lambda x: len(str(x)) == 2 or "-" in str(x)
            )
        ]

        map_arr = []

        # TODO: AAAAAHHHHHH THIS IS SO UGLY
        for code in detailed["6-digit NAICS"]:
            arr = np.zeros(7)
            try:
                series = major[
                    major["6-digit NAICS"]
                    .astype(str)
                    .str.contains(str(code)[:2])
                ]["Protection level"]
                if pd.isnull(series.to_numpy()).any():
                    series = medium[
                        medium["6-digit NAICS"]
                        .astype(str)
                        .str.contains(str(code)[:3])
                    ]["Protection level"]
                    if pd.isnull(series.to_numpy()).any():
                        raise IndexError("Subtract 1 and search again")
                level = list(series)[0]
            except IndexError:
                try:
                    code = int(str(code)[:2]) - 1
                    series = major[
                        major["6-digit NAICS"]
                        .astype(str)
                        .str.contains(str(code))
                    ]["Protection level"]
                    if pd.isnull(series.to_numpy()).any():
                        raise IndexError("Everything failed")
                    level = list(series)[0]
                except IndexError:
                    level = 4.0

            if level % 1 == 0:
                arr[int(level)] = 1
            else:
                arr[math.floor(level)] = 0.5
                arr[math.ceil(level)] = 0.5

            map_arr.append(arr)

        if self.config is None:
            raise ValueError("{self.config=} is null")

        self.level_pm_arr = np.array(map_arr)
        self.level_pm_df = pd.DataFrame(
            self.level_pm_arr,
            index=detailed["Industry description"],
            columns=self.config["Label"]["Demand m"].get(),
        )
        self.level_pm_labs = list(self.level_pm_df.index)
        self.set_description(
            f"{self.level_pm_df=}",
            self.config["Description"]["Population p"]["Protection pm"].get(),
        )
        self.log.debug(f"{self.description['level_pm_df']=}")



"""Get Demand Rates from Dictionary.

The original model based on DOH levels
"""
from typing import Optional

import confuse  # type: ignore

from .demand import Demand  # type: ignore
from .log import Log  # type: ignore
from .organization import Organization  # type: ignore
from .population import Population  # type: ignore
from .resourcemodel import Resource  # type: ignore


class DemandDict(Demand):
    """Calculate demand reading from the data dictionary.

    Overrides the Demand class and most of the calculations are there
    as the default is to use the dictionaries
    """

    def __init__(
        self,
        config: confuse.Configuration,
        res: Resource,
        pop: Optional[Population] = None,
        org: Optional[Organization] = None,
        log_root: Optional[Log] = None,
        type: Optional[str] = None,
    ):
        """Initialize Demand of Resources.

        Calculates the total, costs and per capita demand
        """
        # the init sets all variables from dictionary by default
        super().__init__(config, res, pop, log_root=log_root)
        log = self.log
        log.debug(f"In {__name__}")

        self.pop = pop
        self.res = res
        # These are the core calculations are run wth a single recalc
        self.recalc()



"""The Resource Model.

Reads from dictionary for defaults
"""

import confuse  # type: ignore

from .data import Data  # type: ignore
from .log import Log  # type: ignore
from .resourcemodel import Resource  # type: ignore


class ResourceDict(Resource):
    """ResourceDict - pulls resources from the default config yaml files.

    This can be a default for testing
    """

    def __init__(
        self,
        config: confuse.Configuration,
        log_root: Log = None,
    ):
        """Initialize the resources.

        Reads from the default config yaml files
        """
        # to pick up the description
        super().__init__(config, log_root=log_root)
        log = self.log
        log.debug(f"in {__name__}")

        self.resource_nN_ur = Data("resource_nN_ur", config, log_root=log_root)
        log.debug(f"{self.resource_nN_ur=}")

        self.res_by_popsum1_cost_per_unit_p1n_us = Data(
            "res_by_popsum1_cost_per_unit_p1n_us", config, log_root=log_root
        )
        log.debug(f"{self.res_by_popsum1_cost_per_unit_p1n_us=}")



"""Short script for comparing two CSV files."""
import sys

import pandas as pd  # type: ignore

if __name__ == "__main__":

    # load in the dataframes
    try:
        df1 = pd.read_csv(sys.argv[1])
        df2 = pd.read_csv(sys.argv[2])
    except IndexError:
        print("usage: python validate.py <csv #1> <csv #2>")
        sys.exit()

    # sum all their rows
    sum1 = []
    for col in df1.columns:
        sum1.append(df1[col].sum())
    sum2 = []
    for col in df2.columns:
        sum2.append(df2[col].sum())

    # delete the first row that is just labels from both
    del sum1[0]
    del sum2[0]

    # calculate percent difference
    diff_list = []
    for i, val in enumerate(sum1):
        diff_list.append(float((sum2[i] - val) / val) * 100)

    # display output
    cols = list(df1.columns)[1:]

    print("------------------------")
    print("PERCENT DIFFERENCES")
    for i, val in enumerate(cols):
        print(f"{val}: {diff_list[i]}%")
    print("------------------------")



"""Output module.

Automatically generates config files
"""
from typing import Callable, Dict

import confuse  # type: ignore
import yaml

from .base import Base  # type: ignore
from .log import Log  # type: ignore


class Output(Base):
    """Creates an output object.

    Supply a callback that provides each model element in turn since we
    we cannot pass it. For while mutual import fails so instead we need
    a callback so that output and run through all the elements of model.
    http://effbot.org/pyfaq/how-can-i-have-modules-that-mutually-import-each-other.htm
    """

    def __init__(
        self,
        get_element: Callable,
        config: confuse.Configuration,
        log_root: Log = None,
        out: str = None,
        csv: str = None,
    ):
        """Do the initializing.

        Generate the config files
        """
        super().__init__(log_root=log_root)
        log = self.log
        log.debug(f"In {__name__}")

        self.config = config
        self.out = out
        self.csv = csv

        self.generate_config(get_element)
        self.write_csv()

    def generate_config(self, get_element: Callable):
        """Generate a new config yaml."""
        log = self.log
        new_config: Dict = {}

        for section in ["Parameter", "Dimension", "Paths", "Model"]:
            new_config[section] = self.config[section].get()
        log.debug(f"Wrote all config data into {new_config=}")

        # get the in memory data read from csv's and elsewhere or computed
        # note you cannot pass the model over so you need to do this as a call
        # back but the code looks like
        for key, value in get_element():
            # we have more in memory objects than in the Model
            # so check for existance. For instance filtering is in memory only
            if key in new_config["Model"]:
                new_config["Model"][key]["array"] = value.array

        self.write_config(new_config)

    def write_config(self, config: Dict):
        """Writes config dict to yaml file."""
        if self.out is not None:
            with open(self.out, "w") as yamlfile:
                yaml.dump(config, yamlfile)

    def write_csv(self):
        """Writes to a CSV file."""
        if self.csv is not None:
            df = self.demand.total_demand_pn_df.copy()
            # insert population into the dataframe
            pop = list(self.pop.detail_pd_df["Size"])
            df.insert(loc=0, column="Size", value=pop)
            df.to_csv(self.csv)



"""Setuptools info for the model pip package.

Not yet tested
"""
# https://micropyramid.com/blog/publishing-python-modules-with-pip-via-pypi/
import setuptools  # type:ignore # noqa:

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="restart",
    version="2.6.7",
    author="Restart Partners",
    author_email="lucas@restart.us",
    description="COVID-19 decision model tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/restartus/restart",
    packages=setuptools.find_packages(include=["restart"]),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
)



"""Base for all Classes.

Base mainly includes the description fields
"""
import logging
from typing import Dict, Optional, Tuple

import pandas as pd  # type:ignore

from .log import Log  # type: ignore


class BaseLog:
    """Logging is at the very bottom."""

    def __init__(self, log_root: Optional[Log] = None):
        """Set the Root Log."""
        # since we have no log otherwise
        self.log_root = log_root
        self.log = (
            log_root.log_class(self)
            if log_root is not None
            else logging.getLogger(__name__)
        )
        self.log.debug(f"{self=}")


class Base(BaseLog):
    """Base for all model classes.

    Base strings and description.
    """

    # do not put variable here unless you want them the same
    # across all classes see https://docs.python.org/3/tutorial/classes.html

    # https://stackoverflow.com/questions/9056957/correct-way-to-define-class-variables-in-python
    def __init__(self, log_root: Log = None):
        """Set base varabiles.

        Mainly the descriptions
        """
        super().__init__(log_root=log_root)
        self.description: Dict = {}

    def set_description(self, name: str, description: str):
        """Set the variable description.

        The descriptions are carried in each class so they are self documenting
        May change this to centralized at some point.
        Gets rid of the equal sign if it is there from a f string
        Also only uses the last member name
        """
        # we can't use a higher level logger
        log: logging.Logger = logging.getLogger(__name__)

        # https://stackoverflow.com/questions/18425225/getting-the-name-of-a-variable-as-a-string/58451182#58451182
        # Using Python 3.8 f strings
        # you must use double quotes inside single quotes for strings
        log.debug(f"{object=}")
        # this doesn't work, we need the real object's name so has to happen in
        # caller
        # name = f'{object=}'.split('=')[0]
        # log.debug(f'set self.description[{name}]')
        # https://stackoverflow.com/questions/521502/how-to-get-the-concrete-class-name-as-a-string
        # pdb.set_trace()
        class_name = self.__class__.__name__
        # https://stackoverflow.com/questions/599953/how-to-remove-the-left-part-of-a-string
        # clean up the name so you only get the basename after the period
        # https://www.tutorialspoint.com/How-to-get-the-last-element-of-a-list-in-Python
        name = name.split("=")[0].split(".")[-1]
        model_name = class_name + "." + name
        log.debug(f"{model_name=} {name=}")
        # log.debug(f'set model.description[{model_name}]')
        self.description[name] = description

        # method chaining
        return self

    def __iter__(self):
        """Iterate over all Pandas DataFrames.

        Uses a list of all frames
        """
        self.df_list = [
            k for k, v in vars(self).items() if isinstance(v, pd.DataFrame)
        ]
        self.df_len = len(self.df_list)
        self.df_index = 0
        return self

    def __next__(self) -> Tuple[str, pd.DataFrame]:
        """Next Pandas DataFrame.

        Iterates through the list of dataframes
        """
        if self.df_index >= self.df_len:
            raise StopIteration
        key = self.df_list[self.df_index]
        value = vars(self)[key]
        self.df_index += 1
        return key, value



"""Restart.us Main Module.

Eventually this will become a class and we will
keep the constants like names as private class variables

https://stackoverflow.com/questions/20309456/call-a-function-from-another-file-in-python

"""

import argparse

# https://stackoverflow.com/questions/47561840/python-how-can-i-separate-functions-of-class-into-multiple-files
# explains that you can split a class into separate files by
# putting these inside the class definition
# http://effbot.org/pyfaq/how-do-i-share-global-variables-across-modules.htm
# Before we move to full modules, just import locally
# https://inventwithpython.com/blog/2012/04/06/stop-using-print-for-debugging-a-5-minute-quickstart-guide-to-pythons-logging-module/
import logging  # noqa:F401
from typing import List

from restart.base import Base

# Use jupyter voila instead
# from dashboard import Dashboard
from restart.log import Log
from restart.model import Model
from restart.util import set_config


class Compose:
    """Compose an entire model together.

    Note this must be rentrant for streamlit to work
    So it detects if it has been called again
    """

    def __init__(self):
        """Bootstrap the whole model creating all objects.

        Bootstrap where each modules successively knows more about the world

        In this version, there is a specific order of creation and implicit
        dependencies so use in this order. The notation we uses suffixes each
        different meanings.

        The major model elements are:
        Real World Objects
        Population - People and their attributes
        Organization - Hospitals, EMTs, etc.
        Resources - PPE, Testing, Lab Capacity,...
        Warehouse - Storing resources

        Transformation Objects: These map from real to real
        Demand - The demand from Populations and Organizations for Resources

        Notation: We use snake_case to bind these together

        Object_<dimension description...>_units_dimensions_type

        For example: this means population object as a total count (vs per
        unit) with dimensions:
        g - geo. These are normally filtered out to get to the right level
        t - time. These are normally filtered so it is a single dimension
        r - range of estimates. by default just a single point estimate but can
            be as many different points as you need with lo, mid, hi typically
        p - population one for each class. Note that each object can have
            different summary levels, so p = p0 and l = p1 so the first level
            summary. We allow an unlimited levels, but a single variable with a
            number indicates how much summarization with 0 being the most
            detailed.
        p1 - a shortcut for the "columns" that describe it for this,
            the last element also called the longer form of this is P0 for the
            first level then P1,... as you summarize more. Because python
            doesn't support the "hat" used in the equations hat{p} becomes a
            capital letter P.

        We normally use the shortest notation but all these are valid
        and should point to the same object. With the most common notation just
        noting the object_units_lastdimension_type.

        The notation used here is not full Latex. Use the
        [README.md](README.tex.md) to see the full notation or the Jupyter
        notebook for the long descriptions.

        Pop_tot_desc_gtrpd_df == Pop_tot_gtrpd_df == Pop_tot_gtrp0phat_df ==
        Population_total_description_gtrp0P0_df ==
        Population_geo_time_range_detail_description_detail_gtrp0P0_dataframe

        Summarization levels. Each of these dimensions has a hierarchy:
        As an example with population, population details pd, but you also have
        a summary level population level l, this is a shorthand for where you
        can map say all the SOC or NAICS codes for healthcare professions or
        companies into healthcare. The normal notation for this is

        """
        # set up the logging
        # This name should *not* be the same as any module name like main
        name = "model"
        self.log_root = Log(name)
        log = self.log = self.log_root.log
        # There is a root logger we cannot shutoff so turn off propagation
        log.propagate = False
        # test that logging works
        #  https://docs.python.org/3/howto/logging-cookbook.html
        self.log_root.test(log)
        log.debug(f"{__name__=}")  # goes to log file
        log.info("hello world")  # goes to console

        parser = self.create_parser()
        args = parser.parse_args()
        log.debug(f"{args=}")

        # move config init here so that it can access args
        # now set_configure is just used to change the base default
        self.config = set_config(args.config)
        # self.config = confuse.Configuration("config")
        # Arguments can use dot notation to change
        # Anything in the yaml file
        # https://confuse.readthedocs.io/en/latest/
        self.config.set_args(args, dots=True)
        self.parameter = self.config["Parameter"]

        # uses method chaining
        self.model = model = (
            Model(name, log_root=self.log_root)
            .set_configure(self.config)
            .set_filter(
                county=args.county, state=args.state, subpop=args.subpop
            )
            .set_population(type=args.population)
            .set_organization(type=args.organization)
            .set_resource(type=args.resource)
            .set_inventory(type=args.inventory)
            .set_demand(type=args.demand)
            .set_financial(type=args.financial)
            .set_mobility(type=args.mobility)
            .set_epi(type=args.epi)
            .set_output(out=args.output, csv=args.csv)
        )
        # run the loader and put everything into a super dictionary
        # To change the model, just replace LoadYAML and the configuration
        # of it which starts off the entire model

        # http://net-informations.com/python/iq/instance.htm
        log.debug(f"{model=} is {vars(model)=}")
        for name, value in vars(model).items():
            # http://effbot.org/pyfaq/how-do-i-check-if-an-object-is-an-instance-of-a-given-class-or-of-a-subclass-of-it.htm
            # if issubclass(value, Base):
            if isinstance(value, Base):
                log.debug(f"object {name} holds {value} subclass of Base")

        # test iteration
        for base_key, base_value in model:
            log.debug(f"{base_key=}")
            log.debug(f"{base_value=}")
            for df_key, df_value in base_value:
                log.debug(f"{df_key=}")
                log.debug(f"{df_value=}")

        # TODO: when the rest of the world uses ranges we don't need to this
        # but we need rp1n not p1n, so just stack this
        log.debug("setting average orders")
        model.inventory.set_average_orders_per_period(
            model.demand.demand_by_popsum1_total_rp1n_tc
        )
        log.debug(f"{model.demand.demand_by_popsum1_total_rp1n_tc.df=}")
        # this just keeps increasing supply also test decreasing
        # https://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks
        # fill up the entire range no matter how there are
        backstop_list: List = self.parameter["backstop"].get()
        range_len = model.inventory.inv_by_popsum1_total_rp1n_tc.array.shape[0]
        for index in range(0, len(backstop_list), range_len):
            # note on the :, Black and flake8 are fighting
            backstop = backstop_list[index : index + range_len]  # noqa:E203
            log.critical(f"changing days of backstop to {backstop=}")
            log.debug("reset inventory to zero by ordering everything")
            model.inventory.order(model.inventory.inv_by_popsum1_total_rp1n_tc)
            model.inventory.set_min_in_periods(backstop)
            log.debug(f"{model.inventory.inv_by_popsum1_total_rp1n_tc.df=}")

        # run with streamlit run and then this will not return until after
        # when run as just regular python this doesn't do anything
        log.info("start dashboard ")
        log.debug("start dashboard")

    def create_parser(self):
        """Set Parser arguments.

        For all the choices, returns an argparser object for use by confuse
        """
        # TODO: remove all defaults and set them in the config file
        parser = argparse.ArgumentParser()

        parser.add_argument(
            "-p",
            "--population",
            choices=["dict", "oes", "wa"],
            default="dict",
            help="Select population data cube",
        )

        parser.add_argument("--organization", help=["dict", "ca"])

        parser.add_argument("--csv", help="Select CSV file output")

        parser.add_argument("--county", help="Select county")

        parser.add_argument("--state", help="Select state")

        parser.add_argument(
            "--subpop", help="Select subpopulation of interest"
        )

        parser.add_argument(
            "--config",
            default="restart/config/default",
            help="Select path to config.yaml",
        )

        parser.add_argument("--output", "-o", help="Write results to CSV file")

        parser.add_argument(
            "-r",
            "--resource",
            choices=["dict", "who", "eoc", "chelsea"],
            default="dict",
            help="Select Resource model",
        )

        parser.add_argument(
            "-i",
            "--inventory",
            choices=["dict"],
            default="dict",
            help="Select inventory model",
        )

        parser.add_argument(
            "-d",
            "--demand",
            choices=["mitre", "jhu", "population", "organization"],
            default="population",
            help="Select Demand model",
        )
        parser.add_argument(
            "-f",
            "--financial",
            choices=["dict", "ml", "ensemble"],
            default="dict",
            help="Select Financial model",
        )
        parser.add_argument(
            "-m",
            "--mobility",
            choices=["dict", "apple", "google", "ensemble"],
            default="dict",
            help="Select mobility  model",
        )

        parser.add_argument(
            "-e",
            "--epi",  # only ihme works for now
            choices=[
                "dict",
                "ihme",
                "delphi",
                "icl",
                "lanl",
                "sikjalpha",
                "yyg",
                "chensemble",
            ],
            default="dict",
            help="Select Epi Model",
        )

        return parser

    '''
    # confuse breaks the old model so need to go and fix this
    # deprecated
    def old_compose(self, name, log_root: Optional[Log] = None):
        """Old Model invocation.

        Old Model composition in parts not using chaining
        """
        # https://www.tutorialspoint.com/Explain-Python-class-method-chaining
        if log_root is not None:
            log = log_root.log
        else:
            log = logging.getLogger(__name__)

        log.info("creating Model")
        model: Model = Model(name, log_root=log_root)
        # run the loader and put everything into a super dictionary
        # To change the model, just replace LoadYAML and the configuration
        # of it which starts off the entire model
        loaded = LoadYAML(Path("washington").absolute(), log_root=log_root,)
        log.debug(f"{loaded.data=}")
        log.info("configure Model")
        model.set_configure(loaded)
        # note we cannot just past model down to allow chaining to work
        log.info("creating Population")
        model.population = PopulationDict(
            model.data,
            model.label,
            index="Population p",
            columns="Pop Detail d",
            log_root=log_root,
        )
        log.debug("creating Resource")
        model.resource = Resource(model.data, log_root=model.log_root)
        log.debug("creating Demand")
        model.demand = Demand(model.data, log_root=model.log_root)
        # log.debug("creating Filter")
        # model.filter = Filter(model.data, log_root=model.log_root)
        log.debug("creating Financial")
        model.financial = Financial(model.data, log_root=model.log_root)
        log.debug("creating Epi")
        model.epi = Epi(model.data, log_root=model.log_root)
        log.debug("creating Mobility")
        model.mobility = Mobility(model.data, log_root=model.log_root)
        log.debug(f"{model=}")
    '''


if __name__ == "__main__":
    # compose the entire model runs as a class so it is rentrant
    compose = Compose()
    # Abandon dashboard, use voila and vuetify
    # Dashboard(compose.model)



"""Loader CSV.

The loader for the CSV
"""
import os
from typing import Dict, List, Optional

import pandas as pd  # type:ignore

from .load import Load  # type: ignore
from .log import Log  # type: ignore


class LoadCSV(Load):
    """Converts Excel and CSV files into dataframe objects.

    If you give it files with a .xlsx, .xls, or .csv extension, it will read
    their data into a dataframe, and then safe the dataframe as
    a h5 file with extension .h5. If you feed this class
    a h5 file, it will simply pass through this class. This is
    done so that we can minimize the amount of times the Excel/CSV
    data must be processed - for larger files, it can be lengthy.

    Attributes:
        excel_ext: list of extensions attached to excel files
        csv_ext: list of extensions attached to csv files
        data: dictionary containing names of h5 files
    """

    def __init__(
        self,
        source: Dict = None,
        root: str = None,
        log_root: Optional[Log] = None,
        excel_ext: List[str] = [".xlsx", ".xls"],
        csv_ext: List[str] = [".csv"],
    ):
        """Initialize the Loader to read files.

        Reads the files
        """
        # logging setup
        super().__init__(log_root=log_root)
        log = self.log
        log.debug(f"{self.log=} {log=}")
        log.debug(f"module {__name__=}")

        # extensions we check for
        self.excel_ext = excel_ext
        self.csv_ext = csv_ext

        if source is None:
            raise ValueError(f"{source=} should not be None")

        try:
            if source["Root"] is None:
                raise ValueError(f"need root directory in {source=}")
        except KeyError:
            log.debug(f"{source=} invalid config")
            return None

        # read all files in the given root directory
        if root is None:
            files = os.listdir(os.path.abspath(source["Root"]))
            rootdir = source["Root"]
        else:
            files = os.listdir(root)
            rootdir = root

        self.data: Dict = source

        for fname in source:

            # skip root key
            if not fname == "Root":
                path = source[fname]
                log.debug(f"{path=}")

                # split paths into name + extension
                base, ext = os.path.splitext(path)
                fullbase = os.path.join(rootdir, base)

                try:
                    # look for h5 file in rootdir
                    if base + ".h5" in files:
                        log.debug(f"preexisting json found for {base=}")
                        self.data[fname] = base + ".h5"

                    else:
                        log.debug(f"generating h5 file for {base=}")

                        # excel to dataframe
                        if ext in self.excel_ext:
                            log.debug(f"loading {ext=} file")
                            df = pd.read_excel(fullbase + ext)

                        # csv to dataframe
                        elif ext in self.csv_ext:
                            log.debug(f"loading {ext=} file")
                            df = pd.read_csv(fullbase + ext)

                        else:
                            raise ValueError(f"{fname=} extension invalid")

                        # store dataframe and overwrite dictionary input
                        self.store_dataframe(fullbase, df)
                        self.data[fname] = base + ".h5"

                # handle alternate utf encodings
                except UnicodeDecodeError:
                    log.debug(f"loading {ext=} file with ISO-8859-1 encoding")
                    df = pd.read_csv(fullbase + ext, encoding="ISO-8859-1")
                    self.store_dataframe(fullbase, df)
                    self.data[fname] = base + ".h5"

    def store_dataframe(self, name: str, df: pd.DataFrame) -> None:
        """Serializes a dataframe in h5 format.

        Args:
            name: name of the file we want to save

        Returns:
            None
        """
        log = self.log
        name = name + ".h5"
        log.debug(f"{name=}")
        df.to_hdf(name, key="df", mode="w")

        return None
